'''
使用代码读取csv，清洗数据后写入另一个csv
'''
import pandas as pd
from ezetl.utils import get_reader
from ezetl.utils.common_utils import md5


def trans_volume(volume):
    '''
    转换交易量
    :param volume:
    :return:
    '''
    if str(volume).endswith('K'):
        volume = float(volume[:-1]) * 1000
    if str(volume).endswith('M'):
        volume = float(volume[:-1]) * 1000*1000
    if str(volume).endswith('B'):
        volume = float(str(volume)[:-1].replace(',', '')) * 1000 * 1000 * 1000
    if volume == '-':
        volume = 0
    return round(float(volume), 1)


def trans_date(date):
    '''
    转换日期格式
    :return:
    '''
    year = date.split('年')[0]
    month = date.split('年')[1].split('月')[0]
    day = date.split('年')[1].split('月')[1].split('日')[0]
    return "%d-%02d-%02d 00:00:00" % (int(year), int(month), int(day))


def transform(source):
    dic = {
        'time': trans_date(source['日期']),
        'symbol': f'BTC/USD',
        'timeframe': '1d',
        'low': float(str(source['低']).replace(',', '')),
        'high': float(str(source['高']).replace(',', '')),
        'open': float(str(source['开盘']).replace(',', '')),
        'close': float(str(source['收盘']).replace(',', '')),
        'volume': trans_volume(source['交易量'])
    }
    dic['id'] = md5(f"{dic.get('timeframe')}{dic.get('symbol')}{dic.get('time')}")
    return dic


reader_info = {
    'source': {
        "name": "test",
        "type": "file",
        "conn_conf": {
            "path": "../data/btc_history.csv",
        },
        "ext_params": {}
    },
    'model': {
        "name": "test",
        "type": "file_table",
        "model_conf": {},
        "ext_params": {},
        "fields": []
    },
    'extract_info': {
        'batch_size': 100,
        'extract_rules': []
    }
}
flag, reader = get_reader(reader_info)
print(flag, reader)
flag, res = reader.connect()
print(flag, res)
print(reader.get_res_fields())
data_li = []
for flag, read_data in reader.read_batch():
    print(read_data)
    if read_data['code'] == 200:
        write_data = read_data['data']['records']
        for record in write_data:
            data_li.append(transform(record))
df = pd.DataFrame(data_li)
df.to_csv('btc_history_target.csv', index=False)



import json
import datetime
from ezetl.data_models import DataModel
from kafka import KafkaConsumer, KafkaProducer
from ezetl.libs.kafka_utils import fetch_kafka_data_by_page
from ezetl.utils.common_utils import gen_json_response, parse_json


class DateEncoder(json.JSONEncoder):
    """
    自定义类，解决报错：
    TypeError: Object of type 'datetime' is not JSON serializable
    """

    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return obj.strftime('%Y-%m-%d %H:%M:%S')

        elif isinstance(obj, datetime.date):
            return obj.strftime("%Y-%m-%d")

        else:
            return json.JSONEncoder.default(self, obj)


class KafkaTopicModel(DataModel):
    '''
    Kafka
    '''

    def __init__(self, model_info):
        super().__init__(model_info)
        model_conf = self._model.get('model_conf', {})
        self.topic = model_conf.get('name')
        conn_conf = self._source['conn_conf']
        self.bootstrap_servers = conn_conf['bootstrap_servers']
        conn_setting = {'bootstrap_servers': self.bootstrap_servers}
        self.group_id = None
        if 'ext_params' in model_conf:
            ext_params = parse_json(model_conf.get('ext_params'), {})
            for k in ext_params:
                conn_setting[k] = ext_params[k]
        self.err_info = ''
        self.read_type = 'latest'  # 默认从最新开始读
        self.gen_extract_rules()  # 判断是从头读还是从现在开始读
        conn_setting['auto_offset_reset'] = self.read_type
        if self._extract_info and self.topic:
            try:
                if isinstance(self.topic, list):
                    self.consumer = KafkaConsumer(self.topic[0], **conn_setting)
                    self.consumer.subscribe(self.topic)
                else:
                    self.consumer = KafkaConsumer(self.topic, **conn_setting)
            except Exception as e:
                self.err_info = str(e)
        if self._load_info:
            try:
                self.producer = KafkaProducer(
                    bootstrap_servers=self.bootstrap_servers,
                    value_serializer=lambda x: json.dumps(x, cls=DateEncoder).encode('utf-8'),
                    api_version=(0, 10)
                )
            except Exception as e:
                print(e)
                self.err_info = str(e)

    def connect(self):
        '''
        连通性测试
        '''
        try:
            consumer = KafkaConsumer(bootstrap_servers=self.bootstrap_servers)
            topics = consumer.topics()
            if self.topic and self.topic not in topics:
                return False, '连接失败:未找到该topic'
            return True, '连接成功'
        except Exception as e:
            return False, '连接失败:' + str(e)

    def get_res_fields(self):
        '''
        获取字段列表
        '''
        return None

    def get_search_type_list(self):
        '''
        获取可用高级查询类型
        '''
        return [{
            'name': '从头开始读取',
            'value': 'read_earliest',
            "default": f""
        }, {
            'name': '从最新开始读取',
            'value': 'read_latest',
            "default": f""
        }]

    def get_extract_rules(self):
        '''
        获取可筛选项
        :return:
        '''
        rules = []
        return rules

    def gen_extract_rules(self):
        '''
        解析筛选规则
        :return:
        '''
        rules = [i for i in self.extract_rules if i['field'] == 'search_text']
        if rules != []:
            i = rules[0]
            if i['rule'] == 'read_earliest':
                self.read_type = 'earliest'
            else:
                self.read_type = 'latest'

    def read_page(self, page=1, pagesize=1):
        '''
        分页读取数据
        :param page:
        :param pagesize:
        :return:
        '''
        res_data = fetch_kafka_data_by_page(
            page=page,
            pagesize=pagesize,
            bootstrap_servers=self.bootstrap_servers,
            group_id=self.group_id,
            topic=self.topic
        )
        return True, res_data

    def read_batch(self):
        '''
        生成器分批读取数据
        :return:
        '''
        for msg in self.consumer:
            try:
                data = msg.value
                if isinstance(data, bytes):
                    data = data.decode()
                if isinstance(data, str):
                    try:
                        data = json.loads(data)
                    except Exception as e:
                        print(e)
                res_data = {
                    'records': [data],
                    'total': 1
                }
                yield True, gen_json_response(res_data)
            except Exception as e:
                print(e)

    def write(self, res_data):
        '''
        写入数据
        :param res_data: 
        :return: 
        '''
        self.load_type = self._load_info.get('load_type', '')
        if self.load_type not in ['insert']:
            return False, f'写入类型参数错误,不支持类型{self.load_type}'
        records = []
        if isinstance(res_data, list) and res_data != []:
            records = res_data
        if isinstance(res_data, dict):
            if 'records' in res_data and res_data['records'] != []:
                records = res_data['records']
            else:
                records = [res_data]
        try:
            for c in records:
                self.producer.send(self.topic, value=c)
                print('produce', c)
        except Exception as e:
            return False, f'{str(e)[:-100]}'
        return True, records




#-*- coding:utf-8 -*-
from ezetl.utils.common_utils import import_class, trans_dict_to_rules
from ezetl import reader_map, writer_map


def get_reader(model_info):
    '''
    获取reader对象
    :return:
    '''
    if model_info.get('extract_info') and isinstance(model_info['extract_info'].get('extract_rules'), dict):
        # 若抽取规则为dict形式的，转为列表形式，如{'equal[a]': 1} -> [{{'field': 'a', 'rule': 'equal', 'value': 1}]
        model_info['extract_info']['extract_rules'] = trans_dict_to_rules(model_info['extract_info'].get('extract_rules'))
    source_type = model_info['source'].get('type')
    model_type = model_info['model'].get('type')
    key = f"{source_type}:{model_type}"
    model_class = reader_map.get(key)
    if model_class is None:
        return False, '未找到对应模型'
    try:
        DataModel = import_class(model_class)
        data_model = DataModel(model_info)
        return True, data_model
    except Exception as e:
        return False, str(e)


def get_writer(model_info):
    '''
    获取writer对象
    :return:
    '''
    source_type = model_info['source'].get('type')
    model_type = model_info['model'].get('type')
    key = f"{source_type}:{model_type}"
    model_class = writer_map.get(key)
    if model_class is None:
        return False, '未找到对应模型'
    try:
        DataModel = import_class(model_class)
        data_model = DataModel(model_info)
        return True, data_model
    except Exception as e:
        return False, str(e)


def get_res_fields(res_data):
    '''
    获取返回字段列表
    '''
    res_fields = []
    if isinstance(res_data, list) and res_data != []:
        dic = res_data[0]
        if isinstance(dic, dict):
            res_fields = list(dic.keys())
    if isinstance(res_data, dict):
        if 'records' in res_data:
            if res_data['records'] != []:
                res_fields = list(res_data['records'][0].keys())
            else:
                res_fields = []
        else:
            if 'code' in res_data and res_data['code'] != 200:
                return []
            res_fields = list(res_data.keys())
    return list(set(res_fields))




from pyhive import hive
import pandas as pd


class HiveClient(object):
    def __init__(self, **kwargs):
        """
        host
        port
        username
        password
        database
        """
        self._conn = hive.Connection(**kwargs)

    def create_database(self, database):
        '''
        创建数据库
        '''
        sql = f"create database {database}"
        curosr = self._conn.cursor()
        curosr.execute(sql)
        self._conn.close()

    def get_list_database(self):
        '''
        获取数据库列表
        '''
        cursor = self._conn.cursor()
        cursor.execute('show databases')
        result = cursor.fetchall()
        return result

    def get_tables(self):
        '''
        获取数据库中的表
        '''
        cursor = self._conn.cursor()
        cursor.execute('show databases')
        result = cursor.fetchall()
        return result

    def drop_table(self, table):
        '''
        删除数据库中的表
        '''
        sql = f'''drop table if exists {table}'''
        print(sql)
        cursor = self._conn.cursor()
        cursor.execute(sql)
        self._conn.close()

    def execute(self, sql):
        """
        执行sql
        """
        cursor = self._conn.cursor()
        cursor.execute(sql)
        cursor.execute(sql)
        self._conn.close()

    def query(self, sql):
        """
        返回查询结果列表
        时间默认为秒返回
        """
        cursor = self._conn.cursor()
        cursor.execute(sql)
        result = cursor.fetchall()
        return result

    def query_as_df(self, sql):
        """
        返回查询结果列表
        时间默认为秒返回
        """
        cursor = self._conn.cursor()
        cursor.execute(sql)
        result = cursor.fetchall()
        df = pd.DataFrame(result)
        return df






'''
数据库sqlalchemy orm相关工具类
'''
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, String, Integer, Text, SmallInteger, DateTime, TIMESTAMP, Float
from sqlalchemy.dialects.mysql import LONGTEXT
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.automap import automap_base
from sqlalchemy import create_engine, Table, MetaData
from sshtunnel import SSHTunnelForwarder
Base = declarative_base()


class BaseModel(Base):
    '''
    数据库orm模型基类
    '''
    __abstract__ = True


def transColumn(i, db_type='mysql'):
    '''
    转换表字段，归为统一类型
    '''
    if i['type'].startswith('VARCHAR'):
        i['type'] = 'VARCHAR'
    if db_type == 'clickhouse':
        if i['type'].startswith('FixedString'):
            i['type'] = 'FixedString'
        if i['type'].startswith('Int'):
            i['type'] = 'Int'
        if i['type'].startswith('UInt'):
            i['type'] = 'UInt'
        if i['type'].startswith('Enum'):
            i['type'] = 'Enum'
        if i['type'] == 'FLOAT':
            i['type'] = 'Float'
    return i


def getColumn(i, db_type='mysql'):
    '''
    根据字典获取列参数，组成orm模型字段
    :param i:
    :return:
    '''
    if 'nullable' in i.keys():
        i['nullable'] = i['nullable'] == 1
    else:
        i['nullable'] = True
    if 'is_primary_key' in i:
        i['is_primary_key'] = i['is_primary_key'] == 1
    else:
        i['is_primary_key'] = False
    if 'length' not in i:
        i['length'] = '0'
    if 'type' not in i:
        obj = Column(Text, nullable=True)
    elif i['type'] in ['varchar', 'FixedString']:
        obj = Column(String(i['length']), primary_key=i['is_primary_key'], nullable=i['nullable'])
    elif i['type'] in ['text', 'String']:
        obj = Column(Text, nullable=i['nullable'])
    elif i['type'] == 'longtext':
        if db_type == 'mysql':
            obj = Column(LONGTEXT, nullable=i['nullable'])
        else:
            obj = Column(Text, nullable=i['nullable'])
    elif i['type'] in ['int', 'Int']:
        obj = Column(Integer, primary_key=i['is_primary_key'], nullable=i['nullable'])
    elif i['type'] in ['float', 'Float']:
        obj = Column(Float, nullable=i['nullable'])
    elif i['type'] == 'smallint':
        obj = Column(SmallInteger, nullable=i['nullable'])
    elif i['type'] in ['datetime', 'Date', 'DateTime']:
        obj = Column(DateTime, nullable=i['nullable'])
    elif i['type'] in ['timpstamp']:
        obj = Column(TIMESTAMP, nullable=i['nullable'])
    else:
        obj = Column(Text, nullable=True)
    if 'field_name' in i:
        obj.comment = i['field_name']
    return obj


def create_orm_table(obj_db, params, table_args=None):
    '''
    使用orm创建表
    :return:
    '''
    table_name = params['table_name']
    field_arr = params['field_arr']

    class Model(BaseModel):
        __tablename__ = table_name
        if table_args is not None:
            __table_args__ = table_args

    for i in field_arr:
        setattr(Model, i['field_value'], getColumn(i))
    from sqlalchemy.schema import CreateTable
    print(CreateTable(Model.__table__))
    if table_name not in obj_db.table_names():
        Model.__table__.create(obj_db)
    return Model


def get_database_engine(db_info, res_type='engine'):
    '''
    获取数据库链接引擎
    :param db_obj: 数据库模型
    :return:
    '''
    DB_TYPE = db_info.get('type')
    DB_USER = db_info.get('username')
    DB_PWD = db_info.get('password')
    DB_HOST = db_info.get('host')
    DB_PORT = db_info.get('port')
    DB_NAME = db_info.get('database_name')
    CHARSET = db_info.get('charset')
    if db_info.get('use_tunnel'):
        ssh_info = db_info.get('ssh_tunnel')
        print(ssh_info)
        ssh_host = ssh_info['ssh_host']
        ssh_port = ssh_info['ssh_port']
        ssh_user = ssh_info['ssh_user']
        ssh_passwd = ssh_info['ssh_passwd']
        server = SSHTunnelForwarder(
                (ssh_host, int(ssh_port)),
                ssh_username=ssh_user,
                ssh_password=ssh_passwd,
                remote_bind_address=(DB_HOST, int(DB_PORT))
        )
        server.start()
        DB_PORT = str(server.local_bind_port)
        # server.close()
    ENGINE_DICT = {
        'sqlserver': 'mssql+pymssql',
        'oracle': 'oracle+cx_oracle',
        'pgsql': 'postgresql+psycopg2',
        'mysql': 'mysql+pymysql',
        'clickhouse': 'clickhouse',
        'clickhouse_native': 'clickhouse+native'
    }
    db_engine = None
    if DB_TYPE in ENGINE_DICT:
        db_engine_url = f"{ENGINE_DICT[DB_TYPE]}://{DB_USER}:{DB_PWD}@{DB_HOST}:{DB_PORT}/{DB_NAME}"
        if CHARSET:
            db_engine_url += f'?charset={CHARSET}'
        if res_type == 'engine_url':
            return db_engine_url
        db_engine = create_engine(
            db_engine_url,
            encoding='utf-8')
    return db_engine


def get_database_model(table_name, db_engine=None, db_type='mysql'):
    '''
    根据表名反射出database数据库中的表模型
    :param table_name:
    :return:
    '''
    try:
        DBSession = sessionmaker(bind=db_engine)
        session = DBSession()
        if db_type in ['clickhouse']:
            metadata = MetaData()
            table = Table(table_name, metadata, autoload_with=db_engine)
            return session, table
        else:
            metadata = MetaData()
            metadata.reflect(db_engine, only=[table_name])
            Base = automap_base(metadata=metadata)  # 从metadata中生成所有的映射关系为Base
            Base.prepare()  # 设置被映射的类和关系
            print(Base.classes)
            model = getattr(Base.classes, table_name)  # 将表映射到类上
            return session, model
    except Exception as e:
        print(e)
        return False, False








import pyhdfs


class HdfsClient(object):
    def __init__(self, hosts, user_name='root', randomize_hosts=True, timeout=60, max_tries=2, retry_delay=5, requests_session=None, requests_kwargs=None):
        self.hosts = hosts
        self.randomize_hosts = randomize_hosts
        self.user_name = user_name
        self.timeout = timeout
        self.max_tries = max_tries
        self.retry_delay = retry_delay
        self.requests_session = requests_session
        self.requests_kwargs = requests_kwargs

        self.hdfs_client = self.connect_hdfs()

    def connect_hdfs(self):
        hdfs_client = pyhdfs.HdfsClient(hosts=self.hosts, user_name=self.user_name, timeout=self.timeout)
        return hdfs_client

    # 返回指定目录下的所有文件
    def listdir(self, path="/", **kwargs):
        return self.hdfs_client.listdir(path, **kwargs)

    # 返回用户的根目录
    def get_home_directory(self):
        return self.hdfs_client.get_home_directory()

    # 返回可用的namenode节点
    def get_active_namenode(self):
        return self.hdfs_client.get_active_namenode()

    # 从集群上copy到本地
    # local = /data/123.txt
    # remote = /user/root/123.txt
    def copy_to_local(self, local, remote, **kwargs):
        return self.hdfs_client.copy_to_local(local, remote, **kwargs)

    # 从本地上传文件至集群
    def copy_from_local(self, local, remote, **kwargs):
        return self.hdfs_client.copy_from_local(local, remote, **kwargs)

    # 判断目录、文件是否存在
    def is_exist(self, path, **kwargs):
        return self.hdfs_client.exists(path, **kwargs)

    # 返回目录下的所有目录，路径，文件名
    def walk(self, path, **kwargs):
        return list(self.hdfs_client.walk(path, **kwargs))

    # 删除目录、文件
    def delete(self, path):
        if '.' in path:
            return self.hdfs_client.delete(path)  # 删除文件
        else:
            try:
                return self.hdfs_client.delete(path, recursive=True)  # 删除目录  recursive=True
            except:
                return self.hdfs_client.delete(path)  # 删除文件

    # 打开文件
    def open(self, path, **kwargs):
        response = self.hdfs_client.open(path, **kwargs)
        return response.read()

    # 向已有文件中添加数据
    def append(self, path, data, **kwargs):
        self.hdfs_client.append(path, data, **kwargs)

    # 融合两个文件
    def concat(self, target, source, **kwargs):
        self.hdfs_client.concat(target, source, **kwargs)

    # 创建新目录
    def mkdirs(self, path, **kwargs):
        self.hdfs_client.mkdirs(path, **kwargs)

    # 查看文件的校验和(checksum)
    def checksum(self,path,**kwargs):
        return self.hdfs_client.get_file_checksum(path, **kwargs)

    # 查看当前路径的状态(可路径可文件)
    def list_status(self, path, **kwargs):
        return self.hdfs_client.list_status(path, **kwargs)

    # 重命名
    def rename(self, source, target):
        self.hdfs_client.rename(target, source, target)

    # 创建文件
    def create(self, path, data, **kwargs):
        return self.hdfs_client.create(path, data, **kwargs)






'''
使用任务配置字典实时监听mysql binlog数据流，写入目标kafka数据源
'''
from ezetl.etl_task import etl_task_process
task_params = {
    'extract': {
        'source': {
            "name": "任务数据库",
            "type": "mysql",  # 数据源类型 mysql
            "conn_conf": {  # mysql链接信息
                "host": "127.0.0.1",
                "port": 3306,
                "username": "root",
                "password": "123456",
                "database_name": "naivedata"
            },
            "ext_params": {}
        },
        'model': {
            "name": "任务实例表",
            "type": "mysql_binlog",  # 数据模型类型 mysql表
            "model_conf": {  # 数据模型信息，监听task_instance表的写入，更新，删除操作记录
                "listen_tables": ['task_instance'],
                "only_events": ["write", "update", "delete"]
            },
            "ext_params": {},
            "fields": []
        },
        'extract_info': {
            'extract_type': 'flow',  # 处理类型为流式处理
            'batch_size': 1,
            'extract_rules': []
        }
    },
    'process_rules': [  # 转换流程列表及参数
        {
            "code": "gen_records_list",
            "name": "获取内容列表",
            "rule_dict": {
                "fields": ""
            }
        }
    ],
    'load': {
        'source': {
            "name": "测试kafka写入源",
            "type": "kafka",   # 数据源类型 kafka
            "conn_conf": {   # kafka 连接信息
                "bootstrap_servers": "127.0.0.1:9092",
            },
            "ext_params": {}
        },
        'model': {
            "name": "任务实例topic",
            "type": "kafka_topic",  # 数据模型类型 kafka topic
            "model_conf": {  # 模型信息，topic名task_instance
                "name": "task_instance",
            },
            "ext_params": {},
            "fields": []
        },
        'load_info': {
            'load_type': 'insert',
            'only_fields': []
        }
    }
}
etl_task_process(task_params, run_load=True)



'''
使用代码读取kafka，转换后写入influxdb
'''
from ezetl.utils import get_reader, get_writer
from ezetl.utils.common_utils import format_date

reader_info = {
    'source': {
        "name": "test",
        "type": "kafka",
        "conn_conf": {
            "bootstrap_servers": "127.0.0.1:9092",
        },
        "ext_params": {}
    },
    'model': {
        "name": "test",
        "type": "kafka_topic",
        "model_conf": {
            "name": "btc_history",
        },
        "ext_params": {
            "group_id": "111"
        },
        "fields": []
    },
    'extract_info': {
        'batch_size': 1,
        'extract_rules': []
    }
}
flag, reader = get_reader(reader_info)
print(flag, reader)
flag, res = reader.connect()
print(flag, res)
load_info = {
    'source': {
        "name": "test",
        "type": "influxdb",
        "conn_conf": {
            "host": "127.0.0.1",
            "port": 8086,
            "database": "datacenter",
            "username": "admin",
            "password": "123456"
        },
        "ext_params": {}
    },
    'model': {
        "name": "test",
        "type": "influxdb_table",
        "model_conf": {
            "name": "btc_history",
        },
    },
    'load_info': {
        'load_type': 'insert',
        'only_fields': []
    }
}
_, writer = get_writer(load_info)
print(writer)
flag, res = writer.connect()
print(flag, res)
print(writer.get_res_fields())


def transform(dic):
    '''
    转换数据
    :param dic:
    :return:
    '''
    res_dic = {
        'measurement': 'btc_history',
        'tags': {'timeframe': dic['timeframe'], "symbol": dic['symbol']},
        'fields': {
            'open': float(dic['open']),
            'close': float(dic['close']),
            'high': float(dic['high']),
            'low': float(dic['low']),
            'volume': float(dic['volume'])
        },
        'time': format_date(dic['time'], res_type='timestamp') * 1000000000
    }
    return res_dic

for flag, read_data in reader.read_batch():
    if read_data['code'] == 200:
        write_data = read_data['data']['records']
        write_data = [transform(i) for i in write_data]
        flg, res = writer.write(write_data)
        print(flg, res)


'''
使用代码读取mysql，转换后写入neo4j
'''
from ezetl.utils import get_reader, get_writer

reader_info = {
    'source': {
        "name": "test",
        "type": "mysql",
        "conn_conf": {
            "host": "127.0.0.1",
            "port": 3306,
            "username": "root",
            "password": "123456",
            "database_name": "test"
        },
        "ext_params": {}
    },
    'model': {
        "name": "btc_history",
        "type": "mysql_table",
        "model_conf": {
            "name": "btc_history",
            "auth_type": "create,insert"
        },
        "ext_params": {},
        "fields": []
    },
    'extract_info': {
        'batch_size': 100,
        'extract_rules': []
    }
}
flag, reader = get_reader(reader_info)
print(flag, reader)
flag, res = reader.connect()
print(flag, res)
print(reader.get_res_fields())
read_data = reader.read_page()
print(read_data)
load_info = {
    'source': {
        "name": "test",
        "type": "neo4j",
        "conn_conf": {
            "host": "127.0.0.1",
            "port": 7474,
            "username": "neo4j",
            "password": "123456"
        },
        "ext_params": {}
    },
    'model': {
        "name": "btc_history",
        "type": "neo4j_graph",
        "model_conf": {
            "name": "btc_history",
        },
        "ext_params": {},
        "fields": []
    },
    'load_info': {
        'load_type': 'insert',
        'only_fields': []
    }
}
_, writer = get_writer(load_info)
print(writer)
flag, res = writer.connect()
print(flag, res)
for flag, read_data in reader.read_batch():
    print(read_data)
    if read_data['code'] == 200:
        write_data = read_data['data']['records']
        writer.write(write_data)



'''
映射类转换算法，字段映射，值映射
'''
from ezetl.utils.common_utils import format_date, md5, trans_rule_value, parse_json, parse_to_list, trans_value_type


def map_field_names(source_data={}, rule_dict={}, context={}):
    '''
    字段映射
    :param source_data: 原始数据
    :param rule_dict: 配置信息
    [{
        "name": "字段映射",
        "value": "field_map",
        "form_type": "field_map",
        "required": true,
        "default": "{}",
        "tips": ""
    }]
    :return:
    '''
    field_map = parse_json(rule_dict.get('field_map'))
    if not field_map:
        return False, '缺少必填参数：字段映射'
    if isinstance(source_data, list):
        for i in range(len(source_data)):
            if isinstance(source_data[i], dict):
                for k in field_map:
                    if k in source_data[i] and field_map[k] and field_map[k] != k:
                        source_data[i][field_map[k]] = source_data[i][k]
                        source_data[i].pop(k)
    elif isinstance(source_data, dict):
        for k in field_map:
            if k in source_data and field_map[k] and field_map[k] != k:
                source_data[field_map[k]] = source_data[k]
                source_data.pop(k)
    else:
        # dataframe处理
        source_data = source_data.rename(columns=field_map)
    return True, source_data


def map_values(source_data={}, rule_dict={}, context={}):
    '''
    字段值映射
    :param source_data: 原始数据
    :param rule_dict: 配置信息
     [{
        "name": "处理字段",
        "value": "fields",
        "form_type": "select_fields",
        "required": true,
        "default": "",
        "tips": ""
    }, {
        "name": "映射字典",
        "value": "value_map",
        "form_type": "codeEditor",
        "required": true,
        "default": "{}",
        "tips": ""
    }, {
        "name": "处理错误时",
        "value": "if_error",
        "form_type": "RadioGroup",
        "required": true,
        "default": "original",
        "options": [{
            "label": "原始值",
            "value": "original"
        }, {
            "label": "置空",
            "value": "empty"
        }],
        "tips": ""
    }]
    :return:
    '''
    fields = parse_to_list(rule_dict.get('fields'))
    if not fields:
        return False, '缺少必填参数：处理字段'
    value_map = parse_json(rule_dict.get('value_map'))
    if not value_map:
        return False, '缺少必填参数：映射字典'
    if_error = rule_dict.get('if_error', 'original')
    if isinstance(source_data, list):
        for i in range(len(source_data)):
            if isinstance(source_data[i], dict):
                for field in fields:
                    if field in source_data[i]:
                        try:
                            source_data[i][field] = value_map.get('%s' % source_data[i][field], source_data[i][field])
                        except Exception as e:
                            print(e)
    elif isinstance(source_data, dict):
        for field in fields:
            if field in source_data:
                try:
                    source_data[field] = value_map.get('%s' % source_data[field], source_data[field])
                except Exception as e:
                    print(e)
    else:
        # dataframe处理
        for field in fields:
            if field in source_data.columns:
                try:
                    source_data[field] = source_data[field].apply(lambda x: value_map.get(x, x))
                except Exception as e:
                    return False, str(e)[:500]
    return True, source_data


def trans_time_format(source_data={}, rule_dict={}, context={}):
    '''
    转换时间格式
    :param source_data: 原始数据
    :param rule_dict: 配置信息
    [
      {
        "name": "处理字段",
        "value": "fields",
        "form_type": "select_fields",
        "required": true,
        "default": "",
        "tips": ""
      },
      {
        "name": "日期格式",
        "value": "format",
        "form_type": "input",
        "required": true,
        "default": "%Y-%m-%d %H:%M:%S",
        "tips": ""
      }
    ]
    :return:
    '''
    fields = parse_to_list(rule_dict.get('fields', ''))
    time_format = rule_dict.get('format', '%Y-%m-%d %H:%M:%S')
    if not fields:
        return False, '缺少必填参数：处理字段'
    if isinstance(source_data, list):
        for i in range(len(source_data)):
            if isinstance(source_data[i], dict):
                for field in fields:
                    if field in source_data[i]:
                        try:
                            source_data[i][field] = format_date(source_data[i][field], format=time_format)
                        except Exception as e:
                            print(e)
    elif isinstance(source_data, dict):
        for field in fields:
            if field in source_data:
                try:
                    source_data[field] = format_date(source_data[field], format=time_format)
                except Exception as e:
                    print(e)
    else:
        # dataframe处理
        for field in fields:
            if field in source_data.columns:
                try:
                    source_data[field] = source_data[field].apply(lambda x: format_date(x, format=time_format))
                except Exception as e:
                    return False, str(e)[:500]
    return True, source_data


def trans_field_type(source_data={}, rule_dict={}, context={}):
    '''
    转换字段格式
    :param source_data: 原始数据
    :param rule_dict: 配置信息
    [
      {
        "name": "处理字段",
        "value": "fields",
        "form_type": "select_fields",
        "required": true,
        "default": "",
        "tips": ""
      },{
        "name": "转换类型类型",
        "value": "trans_type",
        "form_type": "select",
        "required": true,
        "default": "sum",
        "options": [
            {"label": "字符串", "value": "str"},
            {"label": "整数", "value": "int"},
            {"label": "浮点数", "value": "float"},
            {"label": "日期字符串", "value": "date"},
            {"label": "日期时间", "value": "datetime"},
            {"label": "时间戳", "value": "timestamp"}
        ],
        "tips": ""
    }
    ]
    :return:
    '''
    fields = parse_to_list(rule_dict.get('fields', ''))
    trans_type = rule_dict.get('trans_type', 'string')
    if not fields:
        return False, '缺少必填参数：处理字段'
    if isinstance(source_data, list):
        for i in range(len(source_data)):
            if isinstance(source_data[i], dict):
                for field in fields:
                    if field in source_data[i]:
                        try:
                            source_data[i][field] = trans_value_type(source_data[i][field], trans_type)
                        except Exception as e:
                            print(e)
    elif isinstance(source_data, dict):
        for field in fields:
            if field in source_data:
                try:
                    source_data[field] = trans_value_type(source_data[field], trans_type)
                except Exception as e:
                    print(e)
    else:
        # dataframe处理
        for field in fields:
            if field in source_data.columns:
                try:
                    source_data[field] = source_data[field].apply(lambda x: trans_value_type(x, trans_type))
                except Exception as e:
                    return False, str(e)[:500]
    return True, source_data


def gen_only_id(source_data={}, rule_dict={}, context={}):
    '''
    生成唯一id
    :param source_data: 原始数据
    :param rule_dict: 配置信息
    [{
        "name": "唯一字段列表",
        "value": "only_fields",
        "form_type": "select_fields",
        "required": true,
        "default": "",
        "tips": ""
    },{
        "name": "唯一号字段",
        "value": "output_field",
        "form_type": "input",
        "required": true,
        "default": "_id",
        "tips": ""
    }]
    :return:
    '''
    only_fields = parse_to_list(rule_dict.get('only_fields', ''))
    output_field = rule_dict.get('output_field')
    if only_fields == []:
        return False, '唯一字段列表不能为空'
    if not output_field:
        return False, '输出字段不能为空'

    if isinstance(source_data, list):
        for i in range(len(source_data)):
            if isinstance(source_data[i], dict):
                s = ''
                for field in only_fields:
                    s += str(source_data[i].get(field))
                only_id = md5(s)
                source_data[i][output_field] = only_id
    elif isinstance(source_data, dict):
        s = ''
        for field in only_fields:
            s += str(source_data.get(field))
        only_id = md5(s)
        source_data[output_field] = only_id
    else:
        # dataframe处理
        try:
            source_data[output_field] = ''
            for field in only_fields:
                if field in source_data.columns:
                    source_data[output_field] += source_data[field].apply(lambda x: str(x))
            source_data[output_field] = source_data[output_field].apply(lambda x: md5(x))
        except Exception as e:
            return False, str(e)[:500]
    return True, source_data


def add_field(source_data={}, rule_dict={}, context={}):
    '''
    添加字段
    :param source_data: 原始数据
    :param rule_dict: 配置信息
    [{
        "name": "字段值",
        "value": "field",
        "form_type": "input",
        "required": true,
        "default": "",
        "tips": ""
    },{
        "name": "默认值",
        "value": "default",
        "form_type": "input",
        "required": true,
        "default": "",
        "tips": ""
    }]
    :return:
    '''
    field = rule_dict.get('field')
    default = trans_rule_value(rule_dict.get('default', ''))
    if not field:
        return False, '字段值不能为空'
    if isinstance(source_data, list):
        for i in range(len(source_data)):
            if isinstance(source_data[i], dict):
               if field not in source_data[i]:
                   source_data[i][field] = default
    elif isinstance(source_data, dict):
        if field not in source_data:
            source_data[field] = default
    else:
        # dataframe处理
        if field not in source_data.columns:
            source_data[field] = default
    return True, source_data



'''
内容获取类转换算法
'''
from ezetl.utils.common_utils import trans_rule_value, parse_to_list


def code_transform(source_data=[], rule_dict={}, context={}):
    '''
    自定义代码转换数据
    :param source_data: 原始数据
    :param rule_dict: 配置信息
    [{
        "name": "语言",
        "value": "language",
        "form_type": "select",
        "required": true,
        "default": "python",
        "options": [{"label": "python", "value": "python"}],
        "tips": ""
    },{
        "name": "代码",
        "value": "code",
        "form_type": "codeEditor",
        "required": true,
        "default": "",
        "tips": ""
    }]
    :return:
    '''
    language = rule_dict.get('language', 'python')
    code = trans_rule_value(rule_dict.get('code'))
    if not code:
        return False, '代码不能为空'
    if language == 'python':
        try:
            exec(code, globals())
            # 调用动态代码中transform函数转换数据
            results = transform(source_data)
            return True, results
        except Exception as e:
            return False, str(e)
    return False, '处理失败'


def data_to_df(source_data=[], rule_dict={}, context={}):
    '''
    将数据转为dataframe
    [{
        "name": "dataframe类型",
        "value": "engine",
        "form_type": "select",
        "required": true,
        "default": "pandas",
        "options": [{"label": "pandas", "value": "pandas"},{"label": "xorbits", "value": "xorbits"}],
        "tips": ""
    }]
    '''
    engine = rule_dict.get('engine', 'pandas')
    if isinstance(source_data, dict):
        source_data = [source_data]
    if engine == 'xorbits':
        import xorbits
        xorbits_cluster = rule_dict.get('xorbits_cluster', 'local')
        if xorbits_cluster != 'local':
            xorbits.init(xorbits_cluster)
        import xorbits.pandas as pd
    else:
        import pandas as pd
    df = pd.DataFrame(source_data)
    return True, df


def df_to_data(source_data=[], rule_dict={}, context={}):
    '''
    dataframe转回原始数据
    []
    '''
    try:
        if not isinstance(source_data, list) and not isinstance(source_data, dict):
            data_li = []
            for k, row in source_data.iterrows():
                data_li.append(row.to_dict())
            return True, data_li
        else:
            return True, source_data
    except Exception as e:
        return False, str(e)[:500]


def gen_records_list(source_data=[], rule_dict={}, context={}):
    '''
    解析接口返回内容列表信息
    :param source_data: 原始数据
    :param rule_dict: 配置信息
    [{
        "name": "字段列表",
        "value": "fields",
        "form_type": "select_fields",
        "required": false,
        "default": "",
        "tips": ""
    }]
    :return:
    '''
    output = []
    fields = parse_to_list(rule_dict.get('fields', ''))
    if isinstance(source_data, dict):
        records = source_data['records']
        for c in records:
            if fields == []:
                dic = c
            else:
                dic = {}
                for field in fields:
                    dic[field] = c.get(field, '')
            output.append(dic)
    return True, output


def gen_contents_first(source_data=[], rule_dict={}, context={}):
    '''
    解析接口第一条数据
    :param source_data: 原始数据
    :param rule_dict: 配置信息
    [{
        "name": "字段",
        "value": "fields",
        "form_type": "select_fields",
        "required": true,
        "default": "",
        "tips": ""
    }]
    :return:
    '''
    fields = parse_to_list(rule_dict.get('fields', ''))
    if isinstance(source_data, dict):
        records = source_data.get('records')
    else:
        records = source_data
    if not records or len(records) == 0:
        return False, '获取数据失败'
    dic = {}
    for field in fields:
        dic[field] = records[0].get(field, '')
    output = [
      dic
    ]
    return True, output


def gen_contents_total(source_data=[], rule_dict={}, context={}):
    '''
    获取总数
    :param source_data: 原始数据
    :param rule_dict: 配置信息
    []
    :return:
    '''
    if isinstance(source_data, dict):
        total_num = source_data.get('total')
    else:
        total_num = len(source_data)
    output = [
        {
            'num': total_num
        }
    ]
    return True, output


def gen_es_aggs_buckets(source_data=[], rule_dict={}, context={}):
    '''
    解析es接口返回聚合统计信息
    :param source_data: 原始数据
    :param rule_dict: 配置信息
    [{
        "name": "统计字段",
        "value": "field",
        "form_type": "select_fields",
        "required": true,
        "default": "",
        "tips": ""
    }, {
        "name": "是否包含其它",
        "value": "include_other",
        "form_type": "switch",
        "required": true,
        "default": false,
        "tips": ""
    }]
    :return:
    '''
    output = []
    field = rule_dict.get('field')
    if not field:
        return False, '缺少必填参数：统计字段'
    aggs = source_data.get('aggs', {})
    res_dic = aggs.get(field)
    if not res_dic or 'buckets' not in res_dic:
        return False, '未找到接口统计字段信息'
    for i in res_dic['buckets']:
        output.append({
            "value": i['doc_count'],
            "name": i['key']
        })
    if rule_dict.get('include_other'):
        output.append({
            "value": res_dic['sum_other_doc_count'],
            "name": '其它'
        })
    return True, output


def gen_es_aggs_value(source_data=[], rule_dict={}, context={}):
    '''
    解析聚类统计数值
    :param source_data: 原始数据
    :param rule_dict: 配置信息
    [{
        "name": "字段",
        "value": "field",
        "form_type": "select_field",
        "required": true,
        "default": "",
        "tips": ""
    }]
    :return:
    '''
    field = rule_dict.get('field')
    if not field:
        return False, '缺少必填参数：字段'
    aggs = source_data.get('aggs')
    if not aggs or aggs == {}:
        return False, '获取数据失败'
    count_res = aggs.get(field)
    if 'value' not in count_res:
        return False, '获取数据失败'
    output = [
      {
        "value": count_res['value']
      }
    ]
    return True, output





import random
from ezetl.data_models import DataModel
from ezetl.utils.common_utils import gen_json_response
from pymysqlreplication import BinLogStreamReader
from pymysqlreplication.row_event import (
    DeleteRowsEvent,
    UpdateRowsEvent,
    WriteRowsEvent
)


class MysqlBinlogModel(DataModel):
    '''
    mysql binlog
    '''

    def __init__(self, model_info):
        super().__init__(model_info)
        conn_conf = self._source['conn_conf']
        self.conn_setting = {
            'host': conn_conf['host'],
            'port': conn_conf['port'],
            'user': conn_conf['username'],
            'passwd': conn_conf['password']
        }
        # 监听数据库
        database = conn_conf.get('database_name')
        self.listen_dbs = [database]
        model_conf = self._model['model_conf']
        # 监听数据表
        self.listen_tables = model_conf.get('listen_tables')
        only_events = model_conf.get('only_events', '')
        if isinstance(only_events, str) and only_events != '':
            only_events = only_events.split(',')
        self.only_events = []
        for event in only_events:
            if event == 'delete':
                self.only_events.append(DeleteRowsEvent)
            if event == 'write':
                self.only_events.append(WriteRowsEvent)
            if event == 'update':
                self.only_events.append(UpdateRowsEvent)
        self.read_type = 'latest'  # 默认从最新开始读

    def connect(self):
        '''
        连通性测试
        '''
        try:
            stream = BinLogStreamReader(
                connection_settings=self.conn_setting,
                server_id=random.randint(10000, 99999),  # slave标识，唯一
                freeze_schema=True,
                resume_stream=True,  # 从最新事件开始监听
                blocking=False,  # 阻塞等待后续事件
                only_schemas=self.listen_dbs,  # 要监听的数据库
                only_tables=self.listen_tables,  # 要监听的表
                # 设定只监控写操作：增、删、改
                only_events=self.only_events
            )
            for binlogevent in stream:
                print(binlogevent)
            return True, '连接成功'
        except Exception as e:
            return False, str(e)

    def get_res_fields(self):
        '''
        获取字段列表
        '''
        return None

    def get_search_type_list(self):
        '''
        获取可用高级查询类型
        '''
        return [{
            'name': '从头开始读取',
            'value': 'read_earliest',
            "default": f""
        }, {
            'name': '从最新开始读取',
            'value': 'read_latest',
            "default": f""
        }]

    def get_extract_rules(self):
        '''
        获取可筛选项
        :return:
        '''
        rules = []
        return rules

    def gen_extract_rules(self):
        '''
        解析筛选规则
        :return:
        '''
        rules = [i for i in self.extract_rules if i['field'] == 'search_text']
        if rules != []:
            i = rules[0]
            if i['rule'] == 'read_earliest':
                self.read_type = 'earliest'
            else:
                self.read_type = 'latest'

    def read_page(self, page=1, pagesize=20):
        '''
        分页读取数据
        :return:
        '''
        stream = BinLogStreamReader(
            connection_settings=self.conn_setting,
            server_id=random.randint(10000, 99999),  # slave标识，唯一
            freeze_schema=True,
            resume_stream=False,  # 预览和分页查询模式，从最开始获取一条数据就返回
            blocking=True,  # 阻塞等待后续事件
            only_schemas=self.listen_dbs,  # 要监听的数据库
            only_tables=self.listen_tables,  # 要监听的表
            # 设定只监控写操作：增、删、改
            only_events=self.only_events
        )
        for binlogevent in stream:
            for row in binlogevent.rows:
                event = {"schema": binlogevent.schema, "table": binlogevent.table}
                if isinstance(binlogevent, DeleteRowsEvent):
                    event["action"] = "delete"
                    event["data"] = row["values"]
                elif isinstance(binlogevent, UpdateRowsEvent):
                    event["action"] = "update"
                    event["data"] = row["after_values"]
                elif isinstance(binlogevent, WriteRowsEvent):
                    event["action"] = "insert"
                    event["data"] = row["values"]
                res_data = {
                    'records': [event],
                    'total': 1
                }
                return True, gen_json_response(data=res_data)

    def read_batch(self):
        '''
        生成器分批读取数据
        :return:
        '''
        self.gen_extract_rules()
        stream = BinLogStreamReader(
            connection_settings=self.conn_setting,
            server_id=random.randint(10000, 99999),  # slave标识，唯一
            freeze_schema=True,
            resume_stream=self.read_type == 'latest',  # 判断是从头开始还是只接受最新数据
            blocking=True,  # 阻塞等待后续事件
            only_schemas=self.listen_dbs,  # 要监听的数据库
            only_tables=self.listen_tables,  # 要监听的表
            # 设定只监控写操作：增、删、改
            only_events=self.only_events
        )
        for binlogevent in stream:
            for row in binlogevent.rows:
                event = {"schema": binlogevent.schema, "table": binlogevent.table}
                if isinstance(binlogevent, DeleteRowsEvent):
                    event["action"] = "delete"
                    event["data"] = row["values"]
                elif isinstance(binlogevent, UpdateRowsEvent):
                    event["action"] = "update"
                    event["data"] = row["after_values"]
                elif isinstance(binlogevent, WriteRowsEvent):
                    event["action"] = "insert"
                    event["data"] = row["values"]
                res_data = {
                    'records': [event],
                    'total': 1
                }
                yield True, gen_json_response(data=res_data)



from py2neo import NodeMatcher, RelationshipMatcher
from ezetl.data_models import DataModel
from ezetl.utils.common_utils import trans_rule_value, flatten_dict, gen_json_response
from ezetl.libs.n4j import NjClient


class N4jGraphModel(DataModel):

    def __init__(self, model_info):
        super().__init__(model_info)
        conn_conf = self._source['conn_conf']
        model_conf = self._model['model_conf']
        self.label = model_conf.get('name')
        n4j_url = f"http://{conn_conf.get('host')}:{conn_conf.get('port')}"
        self.n4j_client = NjClient(n4j_url, **{'username': conn_conf.get('username'), 'password': conn_conf.get('password')})
        self.node_matcher = NodeMatcher(self.n4j_client._client)
        self.relation_matcher = RelationshipMatcher(self.n4j_client._client)

    def connect(self):
        '''
        连通性测试
        '''
        if self.read_page(pagesize=1):
            return True, '连接成功'
        else:
            return False, '连接失败'

    def delete(self):
        '''
        删除
        '''
        try:
            del_sql = f'MATCH (n:{self.label}) detach delete n '
            results = self.n4j_client.query(del_sql)
            return True, '删除成功'
        except Exception as e:
            return False, str(e)

    def get_res_fields(self):
        '''
        获取字段列表
        '''
        return None

    def get_search_type_list(self):
        '''
        获取可用高级查询类型
        '''
        return []

    def get_extract_rules(self):
        '''
        获取可筛选项
        :return:
        '''
        rules = [{
            'name': '等于',
            'value': 'equal'
          }, {
            'name': '不等于',
            'value': 'f_equal'
          }, {
            'name': '包含',
            'value': 'contain'
          }, {
            'name': '不包含',
            'value': 'f_contain'
          }, {
            'name': '大于',
            'value': 'gt'
          }, {
            'name': '大于等于',
            'value': 'gte'
          }, {
            'name': '小于',
            'value': 'lt'
          }, {
            'name': '小于等于',
            'value': 'lte'
          }
        ]
        return rules

    def gen_extract_rules(self):
        '''
        解析筛选规则
        :return:
        '''
        query = self.node_matcher.match(self.label)
        for i in self.extract_rules:
            field = i.get('field')
            rule = i.get('rule')
            value = i.get('value')
            value = trans_rule_value(value)
            if field and field not in ['sql']:
                if value:
                    if isinstance(value, str):
                        value = f"'{value}'"
                    if rule == 'equal':
                        query = query.where(f"_.{field} = {value}")
                    elif rule == 'f_equal':
                        query = query.where(f"NOT _.{field} = {value}")
                    elif rule == 'gt':
                        query = query.where(f"_.{field} > {value}")
                    elif rule == 'lt':
                        query = query.where(f"_.{field} < {value}")
                    elif rule == 'gte':
                        query = query.where(f"_.{field} >= {value}")
                    elif rule == 'lte':
                        query = query.where(f"_.{field} <= {value}")
                    elif rule in ['contain', 'f_contain']:
                        if isinstance(value, str):
                            value = f"'.*{value[1:-1]}.*'"
                        else:
                            value = f"'.*{value}.*'"
                        query = query.where(f"{'NOT ' if rule == 'f_contain' else ''}_.{field} =~ {value}")
        return True, query

    def trans_obj_to_dict(self, obj):
        '''
        将结果转为字典类型
        '''
        dic = {
            'type': 'node',
            'identity': obj.identity,
            'labels': [],
            "properties": {}
        }
        if obj.labels is not None:
            for la in obj.labels:
                dic["labels"].append(la)
        for k, v in obj.items():
            dic["properties"][k] = v
        return dic

    def read_page(self, page=1, pagesize=20):
        '''
        分页读取数据
        :param page:
        :param pagesize:
        :return:
        '''
        if self.n4j_client is False:
            return False, '数据库链接错误'
        flag, query = self.gen_extract_rules()
        if not flag:
            return False, query
        total = query.count()
        query = query.skip((page-1) * pagesize)
        query = query.limit(pagesize)
        obj_list = query.all()
        data_li = []
        for obj in obj_list:
            dic = self.trans_obj_to_dict(obj)
            # 将参数带入字典
            dic = flatten_dict(dic, 'properties')
            data_li.append(dic)
        res_data = {
            'records': data_li,
            'total': total
        }
        return True, gen_json_response(res_data)

    def read_batch(self):
        '''
        生成器分批读取数据
        :return:
        '''
        if self.n4j_client is False:
            yield False, '数据库链接错误'
        flag, query = self.gen_extract_rules()
        if not flag:
            yield False, query
        total = query.count()
        pagesize = self._extract_info.get('batch_size', 1000)
        total_pages = total // pagesize + 1
        n = 0
        while n < total_pages:
            page = n + 1
            n += 1
            query = query.skip((page - 1) * pagesize)
            query = query.limit(pagesize)
            obj_list = query.all()
            data_li = []
            for obj in obj_list:
                dic = self.trans_obj_to_dict(obj)
                # 将参数带入字典
                dic = flatten_dict(dic, 'properties')
                data_li.append(dic)
            res_data = {
                'records': data_li,
                'total': total
            }
            yield True, gen_json_response(res_data)

    def write(self, res_data):
        self.load_type = self._load_info.get('load_type', '')
        if self.load_type not in ['insert', 'update', 'upsert']:
            return False, '写入类型参数错误'
        self.only_fields = self._load_info.get('only_fields', [])
        records = []
        if isinstance(res_data, list) and res_data != []:
            records = res_data
        if isinstance(res_data, dict):
            if 'records' in res_data and res_data['records'] != []:
                records = res_data['records']
            else:
                records = [res_data]
        try:
            insert_records = []
            if self.load_type == 'insert':
                insert_records = records
            if insert_records != []:
                for record in records:
                    self.n4j_client.create_node([self.label], record)
        except Exception as e:
            return False, f'{str(e)[:100]}'
        return True, res_data


class N4jSqlModel(DataModel):

    def __init__(self, model_info):
        super().__init__(model_info)
        conn_conf = self._source.get('conn_conf')
        model_conf = self._model.get('model_conf', {})
        self.sql = model_conf.get('sql', 'CALL db.labels()')
        self.default_sql = self.sql
        self.auth_types = model_conf.get('auth_type', '').split(',')
        n4j_url = f"http://{conn_conf.get('host')}:{conn_conf.get('port')}"
        self.n4j_client = NjClient(n4j_url,
                                   **{'username': conn_conf.get('username'), 'password': conn_conf.get('password')})

    def connect(self):
        '''
        连通性测试
        '''
        if self.read_page(pagesize=1):
            return True, '连接成功'
        else:
            return False, f'连接失败'

    def get_res_fields(self):
        '''
        获取字段列表
        '''
        flag, res = self.read_page(pagesize=1)
        if flag and res.get('code') == 200:
            records = res['data']['records']
            if records != []:
                record = records[0]
                fields = []
                for k in record:
                    field_dic = {
                        'field_name': k,
                        'field_value': k,
                        'ext_params': {}
                    }
                    fields.append(field_dic)
                return fields
        return []

    def get_search_type_list(self):
        '''
        获取可用高级查询类型
        '''
        return [{
            'name': 'sql',
            'value': 'sql',
            "default": self.sql
          }]

    def get_extract_rules(self):
        '''
        获取可筛选项
        :return:
        '''
        rules = []
        return rules

    def gen_extract_rules(self):
        '''
        解析筛选规则
        :return:
        '''
        sql_rules = [i for i in self.extract_rules if i['field'] == 'sql' and i['rule'] == 'search_text' and i['value']]
        if sql_rules != []:
            self.sql = sql_rules[0].get('value')

    def read_page(self, page=1, pagesize=20):
        '''
        分页读取数据
        :param page:
        :param pagesize:
        :return:
        '''
        self.gen_extract_rules()
        if 'custom_sql' not in self.auth_types and self.sql != self.default_sql:
            return False, '无修改sql权限'
        results = self.n4j_client.query(self.sql)
        total = len(results)
        results = results[(page-1)*pagesize:page*pagesize]
        data_li = [obj for obj in results]
        res_data = {
            'records': data_li,
            'total': total
        }
        return True, gen_json_response(res_data)

    def read_batch(self):
        '''
        生成器分批读取数据
        :return:
        '''
        self.gen_extract_rules()
        if 'custom_sql' not in self.auth_types and self.sql != self.default_sql:
            return False, '无修改sql权限'
        results = self.n4j_client.query(self.sql)
        total = len(results)
        data_li = [obj for obj in results]
        pagesize = self._extract_info.get('batch_size', 1000)
        total_pages = total // pagesize + 1
        n = 0
        while n < total_pages:
            page = n + 1
            n += 1
            li = data_li[(page-1)*pagesize:page*pagesize]
            res_data = {
                'records': li,
                'total': total
            }
            yield True, gen_json_response(res_data)





'''
统计类转换算法
'''
import pandas as pd
from ezetl.utils.common_utils import parse_to_list


def group_agg_count(source_data={}, rule_dict={}, context={}):
    '''
    分组聚合统计
    :param source_data: 原始数据
    :param rule_dict: 配置信息
    :param context: 上下文信息
    [{
        "name": "统计字段",
        "value": "count_field",
        "form_type": "select_field",
        "required": true,
        "default": "",
        "tips": ""
    }, {
        "name": "分组字段列表",
        "value": "group_fields",
        "form_type": "select_fields",
        "required": true,
        "default": "",
        "tips": ""
    }, {
        "name": "统计类型",
        "value": "count_type",
        "form_type": "select",
        "required": true,
        "default": "sum",
        "options": [
            {"label": "数量", "value": "count"},
            {"label": "数值总和", "value": "sum"},
            {"label": "平均值", "value": "mean"},
            {"label": "算数平均值", "value": "median"},
            {"label": "标准差", "value": "std"},
            {"label": "方差", "value": "var"},
            {"label": "最小值", "value": "min"},
            {"label": "最大值", "value": "max"},
            {"label": "积", "value": "prod"},
            {"label": "第一个值", "value": "first"},
            {"label": "最后一个值", "value": "last"}
        ],
        "tips": ""
    }]
    :return:
    '''
    count_field = rule_dict.get('count_field')
    count_type = rule_dict.get('count_type', 'sum')
    group_fields = parse_to_list(rule_dict.get('group_fields', ''))
    if group_fields == []:
        return False, '分组字段列表不能为空'
    if not count_field:
        return False, '统计字段不能为空'
    try:
        if isinstance(source_data, list):
            df = pd.DataFrame(source_data)
        elif isinstance(source_data, dict):
            df = pd.DataFrame([source_data])
        else:
            df = source_data
        # dataframe处理
        group_keys = [i for i in group_fields if i in df.columns]
        agg_df = df.groupby(group_keys, as_index=False).agg({count_field: count_type})
        return True, agg_df
    except Exception as e:
        return False, str(e)[:500]





'''
使用任务配置字典分批读取mysql，内置函数转换数据，写入目标kafka数据源
'''
from ezetl.etl_task import etl_task_process

task_params = {
    'extract': {
        'source': {
            "name": "test",
            "type": "mysql",  # 数据源类型 mysql
            "conn_conf": {  # mysql链接信息
                "host": "127.0.0.1",
                "port": 3306,
                "username": "root",
                "password": "123456",
                "database_name": "test"
            },
            "ext_params": {}
        },
        'model': {
            "name": "btc_history",
            "type": "mysql_table",  # 数据模型类型 mysql表
            "model_conf": {  # 数据模型信息，表名btc_history
                "name": "btc_history",
            },
            "ext_params": {},
            "fields": []
        },
        'extract_info': {
            'batch_size': 100,  # 每批读取数量
            # 查询过滤条件 close >= 20000 and high < 40000
            # 可简化为字典形式 'extract_rules': {"get[close]": 30000, "lt[high]": 40000}
            'extract_rules': [
                {'field': "close", "rule": "gte", "value": 20000},
                {'field': "high", "rule": "lt", "value": 40000}
            ]
        }
    },
    'process_rules': [  # 转换流程列表及参数
        {
            "code": "gen_records_list",
            "name": "获取内容列表",
            "rule_dict": {
                "fields": "time,symbol,close,high"
            }
        }
    ],
    'load': {
        'source': {
            "name": "test",
            "type": "kafka",  # 数据源类型 kafka
            "conn_conf": {  # kafka 连接信息
                "bootstrap_servers": "127.0.0.1:9092",
            },
            "ext_params": {}
        },
        'model': {
            "name": "test",
            "type": "kafka_topic",  # 数据模型类型 kafka topic
            "model_conf": {  # 模型信息，topic名btc_history
                "name": "btc_history",
            },
            "ext_params": {},
            "fields": []
        },
        'load_info': {
            'load_type': 'insert',
            'only_fields': []
        }
    }
}
etl_task_process(task_params, run_load=True)



'''
公共工具类，函数
'''
import hashlib
import os.path
import re
import time
import json
import requests
from dateutil.relativedelta import relativedelta
from dateutil import parser
import datetime
import functools
import io
import uuid
import pandas as pd
import importlib


class Singleton(type):
    '''
    单例模式
    '''
    def __init__(cls, *args, **kwargs):
        cls.__instance = None
        super().__init__(*args, **kwargs)

    def __call__(cls, *args, **kwargs):
        if cls.__instance is None:
            cls.__instance = super().__call__(*args, **kwargs)
        return cls.__instance


def import_class(class_path):
    '''
    导入模块
    :param class_path:
    :return:
    '''
    module_name, class_name = class_path.rsplit('.', 1)
    module = importlib.import_module(module_name)
    return getattr(module, class_name)


def gen_json_response(data=None, code=200, msg='ok', msg_key='msg', code_key='code', data_key='data', extends={}):
    '''
    生成web接口返回json格式数据
    :param data:
    :param code:
    :param msg:
    :return:
    '''
    res_data = {
        code_key: code,
        msg_key: msg,
        **extends
    }
    if data is not None:
        res_data[data_key] = data
    return res_data


def gen_uuid(res_type='source'):
    '''
    生成uuid唯一号
    :return:
    '''
    uid = str(uuid.uuid4())
    if res_type == 'base':
        return uid.replace('-', '')
    return uid


def parse_to_list(s, split_text=','):
    '''
    将字符串数据转换为列表
    :param s:
    :param split_text:
    :return:
    '''
    if s is None:
        return []
    if isinstance(s, str):
        if s == '':
            return []
        else:
            s = s.split(split_text)
    return s


def gen_dict_hash_key(dic):
    '''
    对字典生成 hash key
    '''
    # 按字典序将key排序
    sorted_tag_keys = sorted(list(dic.keys()))
    new_dic = {k: dic[k] for k in sorted_tag_keys}
    return md5(str(new_dic))


def get_now_time(res_type='int'):
    '''
    返回当前时间
    '''
    t = time.time()
    if res_type == 'int':
        return int(t)
    if res_type == 'ns':
        return int(t) * 1000000000
    if res_type == 'datetime':
        # t = timestamp_to_date(t)
        # t = format_date(int(t), res_type='datetime')
        return timestamp_to_date(t)
    return t


def print_run_time(func):
    '''
    打印函数运行时间装饰器
    :param func:
    :return:
    '''
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        start = time.time()
        r = func(*args, **kwargs)
        cost = time.time() - start
        cost = round(cost, 3)
        print(f"{func.__name__}函数执行了{cost}s")
        return r
    return wrapper


def parse_json(source, default=None):
    '''
    转为python对象
    :param source:
    :return:
    '''
    if isinstance(source, bytes):
        source = source.decode()
    if isinstance(source, str):
        try:
            source = json.loads(source)
        except Exception as e:
            print(e)
            source = None
    if source is None:
        return default
    return source


def gen_json_to_dict_code(dic, indent=2):
    '''
    将json字符串转为字典代码字符串
    :param dic:
    :param indent:
    :return:
    '''
    dict_code = json.dumps(dic, indent=indent, ensure_ascii=False)
    dict_code = '\n'.join(['{'] + [' ' * indent + i for i in dict_code.split('\n')[1:-1]] + [' ' * indent + '}'])
    dict_code = dict_code.replace(': true', ': True').replace(': false', ': False').replace(': null', ': None')
    return dict_code


def flatten_dict(dic, key, if_exist='replace'):
    '''
    对字典嵌套字典字段摊平到上级字典，如{'field': {'a': 1, 'b': 'hehe'}, 'c': 2020} -> {'a': 1, 'b': 'hehe', 'c': 2020}
    '''
    dic2 = dic.get(key, {})
    for k in dic2:
        if k not in dic or (k in dic and if_exist == 'replace'):
            dic[k] = dic2[k]
    dic.pop(key)
    return dic


def get_json_value(val):
    '''
    将python变量转为json变量
    :param val:
    :return:
    '''
    value = f"'{val}'" if isinstance(val, str) else val
    if value == True:
        value = 'true'
    elif value == False:
        value = 'false'
    elif value is None:
        value = 'null'
    return value


def parse_data_to_excel(data):
    '''
    将数据转为excel文件
    :param data:
    :return:
    '''
    df = pd.DataFrame(data)
    print(df)
    # 使用字节流存储
    output = io.BytesIO()
    # 保存文件
    df.to_excel(output, index=False)
    # 文件seek位置，从头(0)开始
    output.seek(0)
    return output


def request_url(url, method='get', params={}, headers={}, data={}, json={}, timeout=30, retry=3, proxy=None):
    '''
    请求url
    :param url:
    :param retry:
    :return:
    '''
    req_times = 0
    while req_times < retry:
        try:
            res = requests.request(method, url, headers=headers, params=params, data=data, json=json, stream=True, timeout=timeout, proxies=proxy)
            return res
        except Exception as e:
            print(e)
            req_times += 1
    res = requests.request(method, url, headers=headers, params=params, data=data, stream=True, timeout=timeout)
    return res

def trans_value_type(value, trans_type='str'):
    '''
    转换值类型
    :param value:
    :param trans_type:
    :return:
    '''
    if trans_type == 'str':
        value = str(value)
    elif trans_type == 'int':
        try:
            value = int(value)
        except Exception as e:
            print(e)
    elif trans_type == 'float':
        try:
            value = float(value)
        except Exception as e:
            print(e)
    elif trans_type in ['date', 'datetime', 'timestamp']:
        try:
            value = format_date(value, res_type=trans_type)
        except Exception as e:
            print(e)
    return value


def trans_rule_value(value):
    '''
    转换条件筛选时的值
    :param value:
    :return:
    '''
    if not isinstance(value, str):
        return value
    print(value)
    time_dict = {
        's': 1,
        'm': 60,
        'h': 3600,
        'd': 86400,
        'w': 86400 * 7,
        'M': 86400 * 30,
        'Y': 86400 * 365
    }
    #  以timestamp:开头的，转换时间，如&gt[date]=timestamp:-1d 取date大于当前时间往前倒1天的时间戳
    if value.startswith('timestamp:'):
        t = value[-1]
        now = get_now_time()
        try:
            value = int(value[10:-1].strip())
            value = now + value * time_dict[t]
            return value
        except Exception as e:
            print(e)
    elif value.startswith('timestamp_ms:'):
        t = value[-1]
        now = get_now_time()
        try:
            value = int(value[13:-1].strip())
            value = (now + value * time_dict[t]) * 1000
            return value
        except Exception as e:
            print(e)
    #  以time:开头的，转换时间
    #  如&gt[date]=time:-300s 取date大于当前时间往前倒300秒的数据
    elif value.startswith('time:'):
        t = value[-1]
        now = get_now_time()
        try:
            value = int(value[5:-1].strip())
            value = now + value * time_dict[t]
            value = timestamp_to_date(value)
        except Exception as e:
            print(e)
    #  以date:开头的，转换日期时间
    #  如时间2022-02-08 22:10:50 date: %Y-%m-%d %H 将时间转为当前时间格式后转回时间:2022-02-08 22:01:01
    elif value.startswith('date:'):
        now = datetime.datetime.now()
        try:
            value = format_date(now, value[5:].strip())
            value = format_date(value)
        except Exception as e:
            print(e)
    # 以int: 开头的转为整数
    elif value.startswith('int:'):
        try:
            value = int(value[4:])
        except Exception as e:
            print(e)
    # 以float: 开头的转为整数
    elif value.startswith('int:'):
        try:
            value = int(value[6:])
        except Exception as e:
            print(e)
    return value


def get_now_date():
    '''
    获取当前时间
    :return:
    '''
    return str(datetime.datetime.now)[:19]


def format_date(date, format='%Y-%m-%d %H:%M:%S', res_type='str', default=None):
    '''
    格式化日期
    :param date:
    :return:
    '''
    try:
        if isinstance(date, int):
            date = timestamp_to_date(date)
            date = parser.parse(date)
        if isinstance(date, str):
            date = date.replace('年', '-').replace('月', '-').replace('日', ' ').replace('时', ':').replace('分', ':')
            date = parser.parse(date)
        elif not isinstance(date, datetime.datetime):
            date = str(date)
            date = parser.parse(date)
        date = date.strftime(format)
        if res_type == 'date':
            return date
        if res_type == 'datetime':
            date = datetime.datetime.strptime(date, format)
            return date
        if res_type == 'timestamp':
            return date_to_timestamp(date)
        date = str(date)
        if len(date) > 19:
            date = date[:19]
        if '.000Z' in date:
            date = date.replace('.000Z', '')
        if 'T' in date:
            date = date.replace('T', ' ')
        return date
    except Exception as e:
        print(e)
        return default


def date_to_timestamp(date, default=None):
    '''
    日期转时间戳
    :param date:
    :return:
    '''
    try:
        if not isinstance(date, str):
            date = str(date)
        if len(date) > 19:
            date = date[:19]
        if 'T' in date:
            date = date.replace('T', ' ').replace('.000Z', '')
        if len(date) == 10:
            timeArray = time.strptime(date, "%Y-%m-%d")
        else:
            timeArray = time.strptime(date, "%Y-%m-%d %H:%M:%S")
        timeStamp = int(time.mktime(timeArray))
        return timeStamp
    except Exception as e:
        print(e)
        return default


def timestamp_to_date(timestamp, defalut=None):
    '''
    时间戳转日期
    :param date:
    :return:
    '''
    try:
        # 将纳秒转为秒
        if timestamp > 1000000000 * 1000 * 1000:
            timestamp = timestamp / 1000 / 1000
        # 将豪秒转为秒
        if timestamp > 1000000000 * 1000:
            timestamp = timestamp / 1000
        if timestamp != '' and timestamp is not None:
            timeArray = time.localtime(timestamp)
            otherStyleTime = time.strftime("%Y-%m-%d %H:%M:%S", timeArray)
            return otherStyleTime
        else:
            return defalut
    except Exception as e:
        print(e)
        return defalut


def trans_time_length(time_length):
    '''
    转换时间长度，转为秒数
    :param time_length:
    :return:
    '''
    time_symbol_dict = {
        'Y': 86400 * 365,
        'M': 86400 * 30,
        'W': 86400 * 7,
        'd': 86400,
        'h': 3600,
        'm': 60,
        's': 1
    }
    if time_length == 'forever':
        return 86400 * 365 * 100
    else:
        symbol = time_length[-1]
        length = time_length[:-1]
        symbol_len = time_symbol_dict.get(symbol)
        return int(length) * symbol_len


def get_date_list(start_date, end_date):
    '''
    获取日期列表
    '''
    start_time = start_date
    end_time = end_date
    datestart = datetime.datetime.strptime(start_time, '%Y-%m-%d')
    dateend = datetime.datetime.strptime(end_time, '%Y-%m-%d')
    date_list = []
    date_list.append(datestart.strftime('%Y-%m-%d'))
    while datestart < dateend:
        datestart += relativedelta(days=+1)
        date_list.append(datestart.strftime('%Y-%m-%d'))
    query_list = []
    for i in range(len(date_list)):
        s = date_list[i]
        query_list.append(s)
    return query_list


def md5(content):
    '''
    将内容转为MD5值
    :return:
    '''
    md = hashlib.md5()
    if isinstance(content, bytes):
        md.update(content)
    else:
        md.update(content.encode("utf8"))
    return md.hexdigest()


def sha256(content):
    '''
    将内容转为sha256
    :param content:
    :return:
    '''
    s = hashlib.sha256()
    if isinstance(content, bytes):
        s.update(content)
    else:
        s.update(content.encode("utf8"))
    return s.hexdigest()


def parse_to_string(val):
    '''
    将字段或列表转换为json字符串
    :param val:
    :return:
    '''
    if isinstance(val, dict) or isinstance(val, list):
        return json.dumps(val)
    return val


def parse_to_int(val, default=0):
    '''
    将字段或列表转换为数字
    :param val:
    :return:
    '''
    try:
        val = int(val)
    except Exception as e:
        print(e)
        val = default
    return val


def format_str(str1='', type='lower'):
    '''
    去除特殊字符
    :param str1:
    :param type:
    :return:
    '''
    if str1:
        try:
            temp = str1.decode('utf-8', 'ignore')
        except BaseException:
            temp = str1
        xx = u"([\u4e00-\u9fa5a-zA-Z0-9]+)"
        pattern = re.compile(xx)
        results = pattern.findall(str(temp))
        if results:
            long_str = ''.join(results)
            if type == 'lower':
                return long_str.lower()
            elif type == 'upper':
                return long_str.upper()
            else:
                return long_str
        else:
            return ''
    else:
        return ''


def _change_to_lower(*args):
    '''
    转小写
    :param args:
    :return:
    '''
    return str(args[0]).lower()


def _change_to_upper(*args):
    '''
    转大写
    :param args:
    :return:
    '''
    return str(args[0]).upper()


def read_file(file):
    '''
    读取文件转为文件对象
    '''
    file_obj = None
    if isinstance(file, str):
        if file.startswith('http:'):
            try:
                res = request_url(file)
                file_obj = io.BytesIO(res.content)
            except Exception as e:
                print(e)
        else:
            try:
                file_obj = open(file, 'rb')
            except Exception as e:
                print(e)
    return file_obj


def read_file_path(file):
    '''
    读取文件转为文件路径，网络文件生成临时文件
    '''
    if not isinstance(file, str):
        file = str(file)
    if file.startswith('http:'):
        try:
            suffix = file.split('.')[-1]
            res = request_url(file)
            if res.status_code == 200:
                tmp_path = f'{gen_uuid()}.{suffix}'
                tmp_file = open(tmp_path, 'wb')
                tmp_file.write(res.content)
                return True, True, tmp_path
            else:
                return False, False, f'请求网络文件错误:{res.text}'
        except Exception as e:
            return False, False, f'请求网络文件错误:{e}'
    else:
        if os.path.exists(file):
            return True, False, file
        else:
            return False, False, '文件未找到'


def trans_dict_to_rules(api_form):
    '''
    对dict类型数据转换为筛选规则
    '''
    params_list = []
    for k in api_form:
        params_list.append((k, [api_form[k]]))
    extract_rules = []
    for k, values in params_list:
        field = ''
        fields = re.findall('\[(.*?)\]', k)
        if fields != ['']:
            field = '.'.join(fields)
        if '[' in k:
            rule = re.findall('(.*?)\[', k)[0]
        else:
            rule = k
        for value in values:
            extract_rules.append({'field': field, 'rule': rule, 'value': value})
    print(extract_rules)
    return extract_rules






from ezetl.data_models import DataModel
from ezetl.utils.common_utils import gen_json_response
from ezetl.libs.hive import HiveClient


class HiveSqlModel(DataModel):

    def __init__(self, model_info):
        super().__init__(model_info)
        self.db_type = self._source.get('type')
        conn_conf = self._source['conn_conf']
        model_conf = self._model.get('model_conf', {})
        self.sql = model_conf.get('sql', 'show tables')
        self.default_sql = self.sql
        self.auth_types = model_conf.get('auth_type', '').split(',')
        self._client = HiveClient(**conn_conf)

    def connect(self):
        '''
        连通性测试
        '''
        try:
            flag, res = self.read_page(pagesize=1)
            if flag:
                return True, '连接成功'
            else:
                return False, '连接失败'
        except Exception as e:
            return False, f'{e}'

    def get_res_fields(self):
        '''
        获取字段列表
        '''
        flag, res = self.read_page(pagesize=1)
        if flag and res.get('code') == 200:
            records = res['data']['records']
            if records != []:
                record = records[0]
                fields = []
                for k in record:
                    field_dic = {
                        'field_name': k,
                        'field_value': k,
                        'ext_params': {}
                    }
                    fields.append(field_dic)
                return fields
        return []

    def get_search_type_list(self):
        '''
        获取可用高级查询类型
        '''
        return [{
            'name': 'sql',
            'value': 'sql',
            "default": self.sql
        }]

    def get_extract_rules(self):
        '''
        获取可筛选项
        :return:
        '''
        rules = []
        return rules

    def gen_extract_rules(self):
        '''
        解析筛选规则
        :return:
        '''
        sql_rules = [i for i in self.extract_rules if i['field'] == 'search_text' and i['rule'] == 'sql' and i['value']]
        if sql_rules != []:
            self.sql = sql_rules[0].get('value')

    def read_page(self, page=1, pagesize=20):
        '''
        分页读取数据
        :param page:
        :param pagesize:
        :return:
        '''
        self.gen_extract_rules()
        if 'custom_sql' not in self.auth_types and self.sql != self.default_sql:
            return False, '无修改sql权限'
        cursor = self.session.execute(self.sql)
        results = cursor.fetchall()
        total = len(results)
        results = results[(page - 1) * pagesize:page * pagesize]
        data_li = [dict(zip(result.keys(), result)) for result in results]
        res_data = {
            'records': data_li,
            'total': total
        }
        return True, gen_json_response(data=res_data)

    def read_batch(self):
        '''
        生成器分批读取数据
        :return:
        '''
        self.gen_extract_rules()
        if 'custom_sql' not in self.auth_types and self.sql != self.default_sql:
            return False, '无修改sql权限'
        cursor = self._client.execute(self.sql)
        results = cursor.fetchall()
        total = len(results)
        data_li = [dict(zip(result.keys(), result)) for result in results]
        pagesize = self._extract_info.get('batch_size', 1000)
        total_pages = total // pagesize + 1
        n = 0
        while n < total_pages:
            page = n + 1
            n += 1
            li = data_li[(page - 1) * pagesize:page * pagesize]
            res_data = {
                'records': li,
                'total': total
            }
            yield True, gen_json_response(data=res_data)


'''
ES 创建，删除，更新索引等操作封装函数
'''


def get_mapping(data_fields):
    '''
    根据字段组成es mapping
    :return:
    '''
    mapping = {}
    mapping['properties'] = {}
    for field in data_fields:
        mapping['properties'][field.field_value] = {}
        mapping['properties'][field.field_value]['type'] = field.field_type
        if field.field_type == 'date':
            mapping['properties'][field.field_value]['format'] = "yyyy-MM-dd HH:mm:ss||yyyy-MM-dd||epoch_millis"
    return mapping


def set_mapping_field(index_name, field, es=None):
    '''
    根据字段设置es mapping
    :param data_id:
    :return:
    '''
    if es is None:
        raise ValueError('缺少es client')
    mapping = {}
    mapping['properties'] = {}
    mapping['properties'][field.field_value] = {}
    mapping['properties'][field.field_value]['type'] = field.field_type
    if field.field_type == 'date':
        mapping['properties'][field.field_value]['format'] = "yyyy-MM-dd HH:mm:ss||yyyy-MM-dd||epoch_millis"
    try:
        print(mapping)
        es.create_mapping(index_name, mapping)
    except Exception as e:
        print(e)


def get_index_mapping(index_name, es=None):
    '''
    获取es mapping
    :param data_id:
    :return:
    '''
    if es is None:
        raise ValueError('缺少es client')
    try:
        mapping = es.get_mapping(index_name)
        return mapping
    except Exception as e:
        print(e)
        return None


def clean_index(index_name, es=None):
    '''
    清空索引
    :return:
    '''
    if es is None:
        raise ValueError('缺少es client')
    if index_name is None:
        res_data = {
            'code': 400,
            'msg': '未找到索引',
        }
        return res_data
    # 删除索引，重建mapping
    res = es.delete_index(index_name)
    print(res, index_name)
    res_data = {
        'code': 200,
        'msg': '清空成功',
    }
    return res_data


def trans_es_field(field, db_type='mysql'):
    '''
    将es字段类型转化为其他数据库字段类型
    :param field:
    :param db_type:
    :return:
    '''
    if db_type == 'mysql':
        if field['type'] in ['keyword', 'text', 'nested']:
            field['type'] = 'text'
        if field['type'] == 'integer':
            field['type'] = 'int'
        if field['type'] == 'date':
            field['type'] = 'datatime'
    return field


def filter_es_api_field(dic):
    '''
    过滤掉接口带的es字段
    :return:
    '''
    dic['_id'] = dic['id']
    dic.pop('id')
    dic.pop('_index')
    dic.pop('_score')
    dic.pop('_type')
    return dic




# encoding: utf-8
"""
@Description: prometheus 操作封装类
"""
import pandas as pd
from prometheus_api_client import PrometheusConnect


class PrometheusClient(object):
    """
    Prometheus 查询封装
    """

    def __init__(self, **kwargs):
        self._client = PrometheusConnect(**kwargs)

    def get_all_metrics(self):
        '''
        获取metric列表
        '''
        return self._client.all_metrics()

    def query(self, p_sql, **kwargs):
        """
        返回查询结果列表
        """
        return self._client.custom_query(query=p_sql)

    def query_metric_value(self, metric_name, label_config, **kwargs):
        """
        返回查询结果列表
        时间默认为秒返回
        """
        res = self._client.get_current_metric_value(metric_name=metric_name, label_config=label_config)
        return res

    def query_as_df(self, p_sql, **kwargs):
        """
        返回查询结果列表
        时间默认为秒返回
        """
        return pd.DataFrame(self.query(p_sql, **kwargs))



'''
influxdb
'''
from ezetl.data_models import DataModel
from ezetl.libs.ixdb import IxClient
from ezetl.utils.common_utils import trans_rule_value, gen_json_response


class InfluxDBTableModel(DataModel):
    '''
    InfluxDB 表
    '''
    def __init__(self, model_info):
        super().__init__(model_info)
        conn_conf = self._source['conn_conf']
        model_conf = self._model['model_conf']
        self.table_name = model_conf.get('name')
        self.db_name = conn_conf.get('database')
        self.ix_client = IxClient(**conn_conf)

    def connect(self):
        '''
        连通性测试
        '''
        flag, res = self.read_page(pagesize=1)
        if flag:
            return True, '连接成功'
        else:
            return False, '连接失败'

    def delete(self):
        '''
        删除索引
        '''
        try:
            self.ix_client._client.drop_measurement(self.table_name)
            return True, '删除成功'
        except Exception as e:
            return False, str(e)

    def get_res_fields(self):
        '''
        获取字段列表
        '''
        time_dic = {
            'field_name': '时间',
            'field_value': 'time',
            'ext_params': {
                'type': 'time',
            }
        }
        res_fields = [time_dic]
        tags_sql = f'show tag keys on "{self.db_name}" from "{self.table_name}"'
        tags_df = self.ix_client.query_as_df(tags_sql)
        tag_fields = [row['tagKey'] for k, row in tags_df.iterrows()]
        for tag_field in tag_fields:
            dic = {
                'field_name': tag_field,
                'field_value': tag_field,
                'ext_params': {
                    'type': 'tag',
                }
            }
            res_fields.append(dic)
        fields_sql = f'show field keys on "{self.db_name}" from "{self.table_name}"'
        fields_df = self.ix_client.query_as_df(fields_sql)
        field_fields = [row['fieldKey'] for k, row in fields_df.iterrows()]
        for field in field_fields:
            dic = {
                'field_name': field,
                'field_value': field,
                'ext_params': {
                    'type': 'field',
                }
            }
            res_fields.append(dic)
        return res_fields

    def get_search_type_list(self):
        '''
        获取可用高级查询类型
        '''
        return []

    def get_extract_rules(self):
        '''
        获取可筛选项
        :return:
        '''
        rules = [{
            'name': '等于',
            'value': 'equal',
        }, {
            'name': '不等于',
            'value': 'f_equal'
        }, {
            'name': '包含',
            'value': 'contain'
        }, {
            'name': '不包含',
            'value': 'f_contain'
        }, {
            'name': '大于',
            'value': 'gt'
        }, {
            'name': '大于等于',
            'value': 'gte'
        }, {
            'name': '小于',
            'value': 'lt'
        }, {
            'name': '小于等于',
            'value': 'lte'
        }, {
            'name': '时间从大到小排序',
            'value': 'sort_desc'
        }, {
            'name': '时间从小到大排序',
            'value': 'sort_asc'
        }
        ]
        return rules

    def gen_extract_rules(self):
        '''
        解析筛选规则
        :return:
        '''
        filter_sql = ""
        condition_list = []
        sort_str = ''
        for i in self.extract_rules:
            field = i.get('field')
            rule = i.get('rule')
            value = i.get('value')
            value = trans_rule_value(value)
            if value:
                if isinstance(value, str):
                    value = f"'{value}'"
                condition_str = ''
                if rule == 'equal':
                    condition_str = f"{field} = {value}"
                elif rule == 'f_equal':
                    condition_str = f"{field} != {value}"
                elif rule == 'gt':
                    condition_str = f"{field} > {value}"
                elif rule == 'lt':
                    condition_str = f"{field} < {value}"
                elif rule == 'gte':
                    condition_str = f"{field} >= {value}"
                elif rule == 'lte':
                    condition_str = f"{field} <= {value}"
                elif rule == 'contain':
                    condition_str = f"{field}=~/.*{value[1:-1] if isinstance(value, str) else value}.*/"
                if condition_str != '':
                    condition_list.append(condition_str)
            else:
                if rule == 'sort_asc':
                    sort_str = 'ORDER BY time asc'
                elif rule == 'sort_desc':
                    sort_str = 'ORDER BY time desc'
        n = 0
        for i in condition_list:
            if n == 0:
                filter_sql += f" where {i}"
            else:
                filter_sql += f" and {i}"
            n += 1
        if sort_str != '':
            filter_sql += f" {sort_str}"
        return True, filter_sql

    def read_page(self, page=1, pagesize=20):
        '''
        分页读取数据
        :param page:
        :param pagesize:
        :return:
        '''
        flag, filter_sql = self.gen_extract_rules()
        if not flag:
            res_data = {
                'code': 400,
                'msg': filter_sql
            }
            return False, res_data
        count_sql = f"select count(*) from {self.table_name} {filter_sql}"
        print(count_sql)
        total_df = self.ix_client.query_as_df(count_sql)
        print(total_df)
        total = 0
        for k, row in total_df.iterrows():
            row = row.to_dict()
            for k in row:
                total = int(row[k])
        query_sql = f"select * from {self.table_name} {filter_sql} limit {pagesize} offset {(page - 1) * pagesize}"
        print(query_sql)
        df = self.ix_client.query_as_df(query_sql)
        df.fillna("", inplace=True)
        results = []
        for k, row in df.iterrows():
            dic = row.to_dict()
            results.append(dic)
        res_data = {
            'records': results,
            'total': total
        }
        return True, gen_json_response(res_data)

    def read_batch(self):
        '''
        生成器分批读取数据
        :return:
        '''
        flag, filter_sql = self.gen_extract_rules()
        if not flag:
            res_data = {
                'code': 400,
                'msg': filter_sql
            }
            return False, res_data
        count_sql = f"select count(*) from {self.table_name} {filter_sql}"
        print(count_sql)
        total_df = self.ix_client.query_as_df(count_sql)
        print(total_df)
        total = 0
        for k, row in total_df.iterrows():
            row = row.to_dict()
            for k in row:
                total = int(row[k])
        pagesize = self._extract_info.get('batch_size', 1000)
        total_pages = total // pagesize + 1
        n = 0
        while n < total_pages:
            page = n + 1
            n += 1
            query_sql = f"select * from {self.table_name} {filter_sql} limit {pagesize} offset {(page - 1) * pagesize}"
            print(query_sql)
            df = self.ix_client.query_as_df(query_sql)
            results = []
            for k, row in df.iterrows():
                dic = row.to_dict()
                results.append(dic)
            res_data = {
                'records': results,
                'total': total
            }
            yield True, gen_json_response(res_data)

    def write(self, res_data):
        self.load_type = self._load_info.get('load_type', '')
        if self.load_type not in ['insert']:
            return False, f'写入类型参数错误,不支持类型{self.load_type}'
        records = []
        if isinstance(res_data, list) and res_data != []:
            records = res_data
        if isinstance(res_data, dict):
            if 'records' in res_data and res_data['records'] != []:
                records = res_data['records']
            else:
                records = [res_data]
        try:
            if self.load_type == 'insert':
                insert_records = records
                self.ix_client.write_points(insert_records)
        except Exception as e:
            return False, f'{str(e)[:100]}'
        return True, res_data


class InfluxDBSqlModel(DataModel):

    def __init__(self, model_info):
        super().__init__(model_info)
        conn_conf = self._source['conn_conf']
        model_conf = self._model.get('model_conf', {})
        self.sql = model_conf.get('sql', 'show measurements')
        self.default_sql = self.sql
        self.auth_types = model_conf.get('auth_type', '').split(',')
        self.ix_client = IxClient(**conn_conf)

    def connect(self):
        '''
        连通性测试
        '''
        try:
            flag, res = self.read_page(pagesize=1)
            if flag:
                return True, '连接成功'
            else:
                return False, '连接失败'
        except Exception as e:
            return False, f'连接失败:{e}'

    def get_res_fields(self):
        '''
        获取字段列表
        '''
        flag, res = self.read_page(pagesize=1)
        if flag and res.get('code') == 200:
            records = res['data']['records']
            if records != []:
                record = records[0]
                fields = []
                for k in record:
                    field_dic = {
                        'field_name': k,
                        'field_value': k,
                        'ext_params': {}
                    }
                    fields.append(field_dic)
                return fields
        return []

    def get_search_type_list(self):
        '''
        获取可用高级查询类型
        '''
        return [{
            'name': 'sql',
            'value': 'sql',
            "default": self.sql
          }]

    def get_extract_rules(self):
        '''
        获取可筛选项
        :return:
        '''
        rules = []
        return rules

    def gen_extract_rules(self):
        '''
        解析筛选规则
        :return:
        '''
        sql_rules = [i for i in self.extract_rules if i['field'] == 'search_text' and i['rule'] == 'sql' and i['value']]
        if sql_rules != []:
            self.sql = sql_rules[0].get('value')

    def read_page(self, page=1, pagesize=20):
        '''
        分页读取数据
        :param page:
        :param pagesize:
        :return:
        '''
        self.gen_extract_rules()
        if 'custom_sql' not in self.auth_types and self.sql != self.default_sql:
            return False, '无修改sql权限'
        df = self.ix_client.query_as_df(self.sql)
        df.fillna("", inplace=True)
        results = []
        for k, row in df.iterrows():
            results.append(row.to_dict())
        total = len(results)
        results = results[(page-1)*pagesize:page*pagesize]
        res_data = {
            'records': results,
            'total': total
        }
        return True, gen_json_response(data=res_data)

    def read_batch(self):
        '''
        生成器分批读取数据
        :return:
        '''
        self.gen_extract_rules()
        if 'custom_sql' not in self.auth_types and self.sql != self.default_sql:
            return False, '无修改sql权限'
        df = self.ix_client.query_as_df(self.sql)
        df.fillna("", inplace=True)
        results = []
        for k, row in df.iterrows():
            results.append(row.to_dict())
        total = len(results)
        pagesize = self._extract_info.get('batch_size', 1000)
        total_pages = total // pagesize + 1
        n = 0
        while n < total_pages:
            page = n + 1
            n += 1
            li = results[(page-1)*pagesize:page*pagesize]
            res_data = {
                'records': li,
                'total': total
            }
            yield True, gen_json_response(data=res_data)




# encoding: utf-8
"""
@Description: kafka 封装类
"""
from kafka import KafkaConsumer, KafkaProducer
from kafka.structs import TopicPartition


class PrintHandler(object):
    """
    打印处理
    """

    def handle(self, data):
        print(data)


class Consumer(object):
    def __init__(self, *topic, **config):
        """
        'bootstrap_servers': 'localhost',
        'client_id': 'kafka-python-' + __version__,
        'group_id': None,
        """
        self._consumer = KafkaConsumer(*topic, **config)

    def run(self, handler):
        for msg in self._consumer:
            try:
                if handler:
                    # print(msg.value)
                    handler.handle(msg.value)
            except Exception as e:
                print(e)


class Producer(object):
    def __init__(self, topic, **config):
        """
        'bootstrap_servers': 'localhost',
        'client_id': None,
        """
        self._producer = KafkaProducer(**config)
        self._topic = topic

    def send(self, data):
        self._producer.send(self._topic, value=bytes(data, encoding='utf-8'), key=None)
        self._producer.flush()

    def close(self):
        self._producer.close()


def get_newest_offset(bootstrap_servers, topic):
    '''
    获取topic最新offset
    '''
    consumer = KafkaConsumer(bootstrap_servers=bootstrap_servers)
    partitions_set = consumer.partitions_for_topic(topic)
    # 获取最新偏移量
    zz = [TopicPartition(topic, p) for p in partitions_set]
    topic_offset_dict = consumer.end_offsets(zz)
    result = list(topic_offset_dict.values())
    if result != []:
        return max(result)
    else:
        return 0


def get_group_newest_offset(bootstrap_servers, group_id, topic):
    """
    获取一个topic特定group已经消费的offset
    """
    consumer = KafkaConsumer(bootstrap_servers=bootstrap_servers,
                             group_id=group_id,
                             )
    pts = [TopicPartition(topic=topic, partition=i) for i in
           consumer.partitions_for_topic(topic)]
    result = consumer._coordinator.fetch_committed_offsets(pts)
    result = [r.offset for r in result.values()]
    if result != []:
        return max(result)
    else:
        return 0


def fetch_kafka_data_by_page(page=0, pagesize=10, bootstrap_servers=None, topic=None, group_id=None):
    '''
    分页读取kafka数据
    '''
    if bootstrap_servers is None:
        return {
            'code': 400,
            'msg': '链接地址不能为空!'
        }
    elif topic is None:
        return {
            'code': 400,
            'msg': 'topic不能为空!'
        }
    elif group_id is not None:
        consumer = KafkaConsumer(
            group_id=group_id,  # 指定此消费者实例属于的组名，可以不指定
            bootstrap_servers=bootstrap_servers,  # 指定kafka服务器
            auto_offset_reset='earliest'
        )
        total = get_group_newest_offset(bootstrap_servers, group_id, topic)
    else:
        consumer = KafkaConsumer(
            bootstrap_servers=bootstrap_servers,  # 指定kafka服务器
            auto_offset_reset='earliest'
        )
        total = get_newest_offset(bootstrap_servers, topic)
    tp = TopicPartition(topic, 0)
    consumer.assign([tp])
    start_offset = (page-1)*pagesize
    end_offset = start_offset + pagesize
    if end_offset > total:
        return {
            'code': 200,
            'msg': 'ok',
            'records': [],
            'total': total
        }
    data_li = []
    consumer.seek(tp, start_offset)
    n = 0
    for msg in consumer:
        data = msg.value.decode('utf-8').encode('utf-8').decode('unicode_escape')
        data_li.append(data)
        if n >= pagesize:
            return {
                'code': 200,
                'msg': 'ok',
                'records': data_li,
                'total': total
            }
        n += 1




import os.path
from ezetl.data_models import DataModel
from ezetl.utils.common_utils import trans_rule_value, read_file, md5, gen_json_response
import pandas as pd
import io


class BaseFileModel(DataModel):

    def __init__(self, model_info):
        super().__init__(model_info)
        conn_conf = self._source.get('conn_conf', {})
        self.file_path = conn_conf.get('path')

    def connect(self):
        '''
        连通性测试
        '''
        try:
            file_obj = read_file(self.file_path)
            if file_obj:
                return True, '连接成功'
            else:
                return False, '连接失败'
        except Exception as e:
            return False, str(e)


class TableFileModel(BaseFileModel):

    def __init__(self, model_info):
        super().__init__(model_info)

    def connect(self):
        '''
        连通性测试
        '''
        flag, res = self.read_page(pagesize=1)
        if flag:
            return True, '连接成功'
        else:
            return False, '连接失败'

    def get_res_fields(self):
        '''
        获取字段列表
        '''
        flag, res_data = self.gen_extract_rules()
        if not flag:
            assert ValueError(res_data)
        res_fields = []
        try:
            for column in self.df.columns:
                dic = {
                    'field_name': column,
                    'field_value': column,
                }
                res_fields.append(dic)
        except Exception as e:
            print(e)
        return res_fields

    def read_file_path(self, file_path, use_cache=False, cache_size=1000):
        '''
        预览模式读取较大网络文件时缓存文件前n条直接读取
        '''
        if use_cache:
            if not os.path.exists('tmp'):
                os.mkdir('tmp')
            cache_file_path = os.path.join('tmp', f"{md5(file_path)}_{cache_size}")
            if os.path.exists(cache_file_path):
                content_value = open(cache_file_path, 'rb').read()
                file_obj = io.BytesIO(content_value)
                return file_obj
            else:
                file_obj = read_file(file_path)
                df = self.get_df(file_obj, nrows=cache_size)
                f = io.BytesIO()
                if self.file_path.endswith('.csv'):
                    df.to_csv(f, index=False)
                else:
                    df.to_excel(f, index=False)
                content_value = f.getvalue()
                cache_file = open(cache_file_path, 'wb')
                cache_file.write(content_value)
                cache_file.close()
                return f
        else:
            file_obj = read_file(file_path)
            return file_obj

    def get_df(self, file_obj, nrows=None, skiprows=None):
        '''
        获取pandas df
        '''
        df = None
        try:
            if self.file_path.endswith('.csv'):
                df = pd.read_csv(file_obj, dtype=object, nrows=nrows, skiprows=skiprows)
            else:
                df = pd.read_excel(file_obj, dtype=object, nrows=nrows, skiprows=skiprows)
            df.fillna("", inplace=True)
        except Exception as e:
            print(e)
        return df

    def get_search_type_list(self):
        '''
        获取可用高级查询类型
        '''
        return []

    def get_extract_rules(self):
        '''
        获取可筛选项
        :return:
        '''
        rules = [{
            'name': '等于',
            'value': 'equal'
          }, {
            'name': '不等于',
            'value': 'f_equal'
          }, {
            'name': '大于',
            'value': 'gt'
          }, {
            'name': '大于等于',
            'value': 'gte'
          }, {
            'name': '小于',
            'value': 'lt'
          }, {
            'name': '小于等于',
            'value': 'lte'
          }, {
            'name': '缓存前n条数据读取',
            'value': 'use_cache'
          }
        ]
        return rules

    def gen_extract_rules(self):
        '''
        解析筛选规则
        :return:
        '''
        cache_rules = [i for i in self.extract_rules if i.get('rule') == 'use_cache']
        cache_size = 0
        if cache_rules != []:
            try:
                cache_size = int(cache_rules[0].get('value'))
            except Exception as e:
                print(e)
        if cache_size > 0:
            self.file_obj = self.read_file_path(self.file_path, use_cache=True, cache_size=cache_size)
        else:
            self.file_obj = self.read_file_path(self.file_path, use_cache=False)
        self.df = self.get_df(self.file_obj)
        if self.df is None:
            return False, '文件读取错误'
        for i in self.extract_rules:
            print(i)
            field = i.get('field')
            rule = i.get('rule')
            value = i.get('value')
            value = trans_rule_value(value)
            if field and value:
                if rule == 'equal':
                    self.df = self.df[self.df[field] == value]
                elif rule == 'f_equal':
                    self.df = self.df[self.df[field] != value]
                elif rule == 'gt':
                    self.df = self.df[self.df[field] > value]
                elif rule == 'gte':
                    self.df = self.df[self.df[field] >= value]
                elif rule == 'lt':
                    self.df = self.df[self.df[field] < value]
                elif rule == 'lte':
                    self.df = self.df[self.df[field] <= value]
        data_li = []
        for k, row in self.df.iterrows():
            data_li.append(row.to_dict())
        return True, data_li

    def read_page(self, page=1, pagesize=20):
        '''
        分页读取数据
        :param page:
        :param pagesize:
        :return:
        '''
        flag, res_data = self.gen_extract_rules()
        if not flag:
            return False, res_data
        total = len(res_data)
        data_li = res_data[(page-1)*pagesize:page*pagesize]
        res_data = {
            'records': data_li,
            'total': total
        }
        return True, gen_json_response(data=res_data)

    def read_batch(self):
        '''
        生成器分批读取数据
        :param res_type: 返回形式
        :return:
        '''
        flag, res_data = self.gen_extract_rules()
        if not flag:
            return False, res_data
        total = len(res_data)
        pagesize = self._extract_info.get('batch_size', 1000)
        total_pages = total // pagesize + 1
        n = 0
        while n < total_pages:
            page = n + 1
            n += 1
            result = {
                'records': res_data[(page-1)*pagesize:page*pagesize],
                'total': total
            }
            yield True, gen_json_response(result)






# encoding: utf-8
"""
neo4j 操作封装类
"""
from py2neo import Graph, Node, Relationship
import time


class NjClient(object):
    def __init__(self, p_url, **kwargs):
        """
        NjClient('http://localhost:7474', **{'username':'', 'password':''})
        新版本认证使用auth=(kwargs['username'], kwargs['password'])
        """
        try:
            self._client = Graph(p_url, **kwargs)
        except Exception as e:
            print(e)
            self._client = Graph(p_url, auth=(kwargs['username'], kwargs['password']))

    def delete_all(self):
        self._client.delete_all()

    def query(self, p_query):
        return self._client.run(p_query).data()

    def find(self, p_labels, p_properties):
        return self._client.nodes.match(*p_labels, **p_properties)

    def find_one(self, p_labels, p_properties):
        """
        查询节点
        """
        return self.find(p_labels, p_properties).first()

    def create_node(self, p_labels, p_properties):
        """
        创建一个新节点
        """
        self._client.create(Node(*p_labels, **p_properties))
        return self.find_one(p_labels, p_properties)

    def create_node_with_check(self, p_labels, p_check_properties, p_properties):
        """
        查询并创建节点
        p_labels label列表
        p_check_properties 确定节点唯一性属性
        p_properties 创建或者更新的属性列表
        """
        node = self.find_one(p_labels, p_check_properties)
        if node is None:
            return self.create_node(p_labels, p_properties)

        return node

    def update_node(self, p_labels, p_key_properties, p_data_properties):
        """
        更新节点
        p_labels label列表
        p_key_properties 确定节点唯一性属性
        p_properties 创建或者更新的属性列表
        """
        node = self.find_one(p_labels, p_key_properties)
        if node is None:
            return None

        for key, value in p_data_properties.items():
            node[key] = value
        node['update_time'] = time.strftime('%Y-%m-%d %H:%M:%S')
        self._client.push(node)
        return node

    def update_or_create_node(self, p_labels, p_properties):
        """
        更新或者新建节点
        p_labels label列表
        p_properties 确定节点唯一性属性也是更新的属性
        """
        node = self.update_node(p_labels, p_properties, p_properties)
        if node:
            return node

        node = self.create_node(p_labels, p_properties)
        return node

    def update_or_create_node_with_key(self, p_labels, p_key_properties, p_properties):
        """
        更新或者新建节点
        p_labels label列表
        p_key_properties 确定节点唯一性属性
        p_properties 创建或者更新的属性列表
        """
        node = self.update_node(p_labels, p_key_properties, p_properties)
        if node:
            return node

        node = self.create_node(p_labels, p_properties)
        return node

    def update_with_key(self, p_labels, p_key_properties, p_properties):
        """
        更新节点属性
        """
        nodes = self.find(p_labels, p_key_properties)
        for node in nodes.all():
            for key, value in p_properties.items():
                node[key] = value
            self._client.push(node)

    def update_by_identity(self, identity, p_properties):
        """
        根据节点ID更新属性
        """
        node = self._client.nodes.get(identity)
        if node:
            for key, value in p_properties.items():
                node[key] = value
            self._client.push(node)

    def update_by_identity_with_ext(self, identity, p_properties):
        """
        根据节点ID更新属性，同时更新部分附加固有属性
        """
        node = self._client.nodes.get(identity)
        if node:
            for key, value in p_properties.items():
                node[key] = value
            node['gmt_modified'] = int(time.time())
            node['gmt_modified_datetime'] = time.strftime('%Y-%m-%d %H:%M:%S')
            self._client.push(node)

    def find_by_identity(self, identity):
        """
        根据节点ID获取节点
        """
        return self._client.nodes.get(identity)

    def create_relation(self, start_node, label, end_node, relation_info):
        """
        创建节点关系
        """
        relation = Relationship(start_node, label, end_node, **relation_info)
        return self._client.create(relation)





reader_map = {
    'file:None': 'ezetl.data_models.file_models.BaseFileModel',
    'file:file_table': 'ezetl.data_models.file_models.TableFileModel',
    'mysql:None': 'ezetl.data_models.base_db_sql.BaseDBSqlModel',
    'mysql:sql': 'ezetl.data_models.base_db_sql.BaseDBSqlModel',
    'mysql:mysql_table': 'ezetl.data_models.mysql_table.MysqlTableModel',
    'mysql:mysql_binlog': 'ezetl.data_models.mysql_binlog.MysqlBinlogModel',
    'clickhouse:None': 'ezetl.data_models.clickhouse_table.BaseDBSqlModel',
    'clickhouse:sql': 'ezetl.data_models.base_db_sql.BaseDBSqlModel',
    'clickhouse:clickhouse_table': 'ezetl.data_models.clickhouse_table.CkTableModel',
    'elasticsearch:None': 'ezetl.data_models.elasticsearch_index.EsIndexModel',
    'elasticsearch:elasticsearch_index': 'ezetl.data_models.elasticsearch_index.EsIndexModel',
    'mongodb:None': 'ezetl.data_models.mongo_models.MongoModel',
    'mongodb:mongodb_collection': 'ezetl.data_models.mongo_models.MongoModel',
    'neo4j:None': 'ezetl.data_models.neo4j_models.N4jSqlModel',
    'neo4j:sql': 'ezetl.data_models.neo4j_models.N4jSqlModel',
    'neo4j:neo4j_graph': 'ezetl.data_models.neo4j_models.N4jGraphModel',
    'influxdb:None': 'ezetl.data_models.ixdb_models.InfluxDBSqlModel',
    'influxdb:sql': 'ezetl.data_models.ixdb_models.InfluxDBSqlModel',
    'influxdb:influxdb_table': 'ezetl.data_models.ixdb_models.InfluxDBTableModel',
    'kafka:None': 'ezetl.data_models.kafka_topic.KafkaTopicModel',
    'kafka:kafka_topic': 'ezetl.data_models.kafka_topic.KafkaTopicModel'
}

writer_map = {
    'mysql:mysql_table': 'ezetl.data_models.mysql_table.MysqlTableModel',
    'clickhouse:clickhouse_table': 'ezetl.data_models.clickhouse_table.CkTableModel',
    'elasticsearch:elasticsearch_index': 'ezetl.data_models.elasticsearch_index.EsIndexModel',
    'kafka:kafka_topic': 'ezetl.data_models.kafka_topic.KafkaTopicModel',
    'mongodb:mongodb_collection': 'ezetl.data_models.mongo_models.MongoModel',
    'neo4j:neo4j_graph': 'ezetl.data_models.neo4j_models.N4jGraphModel',
    'influxdb:influxdb_table': 'ezetl.data_models.ixdb_models.InfluxDBTableModel',
}



# encoding: utf-8
"""
@Description: influxdb 操作封装类
"""
import time
import pandas as pd
from influxdb import InfluxDBClient


class IxClient(object):
    """
    :param host: hostname to connect to InfluxDB, defaults to 'localhost'
    :type host: str
    :param port: port to connect to InfluxDB, defaults to 8086
    :type port: int
    :param username: user to connect, defaults to 'root'
    :type username: str
    :param password: password of the user, defaults to 'root'
    :type password: str
    :param pool_size: urllib3 connection pool size, defaults to 10.
    :type pool_size: int
    :param database: database name to connect to, defaults to None
    :type database: str
    :param ssl: use https instead of http to connect to InfluxDB, defaults to
        False
    :type ssl: bool
    :param verify_ssl: verify SSL certificates for HTTPS requests, defaults to
        False
    :type verify_ssl: bool
    :param timeout: number of seconds Requests will wait for your client to
        establish a connection, defaults to None
    :type timeout: int
    :param retries: number of retries your client will try before aborting,
        defaults to 3. 0 indicates try until success
    :type retries: int
    :param use_udp: use UDP to connect to InfluxDB, defaults to False
    :type use_udp: bool
    :param udp_port: UDP port to connect to InfluxDB, defaults to 4444
    :type udp_port: int
    :param proxies: HTTP(S) proxy to use for Requests, defaults to {}
    :type proxies: dict
    :param path: path of InfluxDB on the server to connect, defaults to ''
    :type path: str
    :param cert: Path to client certificate information to use for mutual TLS
        authentication. You can specify a local cert to use
        as a single file containing the private key and the certificate, or as
        a tuple of both files’ paths, defaults to None
    :type cert: str
    :param gzip: use gzip content encoding to compress requests
    :type gzip: bool
    :param session: allow for the new client request to use an existing
        requests Session, defaults to None
    :type session: requests.Session
    :raises ValueError: if cert is provided but ssl is disabled (set to False)
    """

    def __init__(self, **kwargs):
        self._client = InfluxDBClient(**kwargs)

    def create_database(self, database):
        '''
        创建数据库
        '''
        self._client.create_database(database)

    def get_list_database(self):
        '''
        获取数据库列表
        '''
        return self._client.get_list_database()

    def get_tables(self):
        '''
        获取数据库中的表
        '''
        return self._client.query("show measurements")

    def query(self, p_sql, **kwargs):
        """
        返回查询结果列表
        时间默认为秒返回
        """
        if not kwargs.get('epoch'):
            kwargs['epoch'] = 's'
        return self._client.query(p_sql, **kwargs)

    def query_as_points(self, p_sql, **kwargs):
        """
        返回查询结果列表
        时间默认为秒返回
        """
        if not kwargs.get('epoch'):
            kwargs['epoch'] = 's'
        return self._client.query(p_sql, **kwargs).get_points()

    def query_as_df(self, p_sql, **kwargs):
        """
        返回查询结果列表
        时间默认为秒返回
        """
        if not kwargs.get('epoch'):
            kwargs['epoch'] = 's'
        return pd.DataFrame(self.query_as_points(p_sql, **kwargs))

    def write_points(self, p_points):
        self._client.write_points(p_points, batch_size=5000)


class IxOutput(object):
    """
    为了提高性能，部分写入操作可以适当缓存写入
    """

    def __init__(self, **kwargs):
        self._client = IxClient(**kwargs)
        self._cache = []
        self._cache_max_size = kwargs.get('cache_size', 500)
        self._cache_max_seconds = kwargs.get('cache_seconds', 2)
        self._cache_last_time = time.time()

    def save_with_cache(self, points):
        """
        缓存写入数据
        """
        self._cache.extend(points)
        if len(self._cache) >= self._cache_max_size or time.time() - self._cache_last_time >= self._cache_max_seconds:
            self._client.write_points(self._cache)
            self._cache.clear()
            self._cache_last_time = time.time()

    def save(self, points):
        self._client.write_points(points)

    def save_when_exit(self):
        if len(self._cache) > 0:
            self._client.write_points(self._cache)
            self._cache.clear()
            self._cache_last_time = time.time()




# encoding: utf-8
"""
elastic search 操作封装类
"""
from elasticsearch import Elasticsearch
from elasticsearch import helpers


class EsClient(object):
    def __init__(self, hosts=None, **kwargs):
        """
                :arg hosts: list of nodes, or a single node, we should connect to.
                    Node should be a dictionary ({"host": "localhost", "port": 9200}),
                    the entire dictionary will be passed to the :class:`~elasticsearch.Connection`
                    class as kwargs, or a string in the format of ``host[:port]`` which will be
                    translated to a dictionary automatically.  If no value is given the
                    :class:`~elasticsearch.Connection` class defaults will be used.
                :arg params: any additional arguments will be passed on to the
                    :class:`~elasticsearch.Transport` class and, subsequently, to the
                    :class:`~elasticsearch.Connection` instances.

                exsample：Elasticsearch(hosts=[{'host':'localhost','port':9200}])
                exsample：Elasticsearch(hosts=['node1','node2'])
                """
        self._client = Elasticsearch(hosts=hosts, **kwargs)

    def base_query(self, **kwargs):
        """
        elastic search 原始查询类，主要入参 index，body，size，headers
        """
        return self._client.search(**kwargs)

    def query(self, p_index, p_query, p_size=1, headers=None):
        """
        常用查询函数
        """
        return self.base_query(index=p_index, body=p_query, size=p_size, headers=headers)

    def query_count(self, p_index, p_query, headers=None):
        """
        常用计数函数
        """
        l_result = self.base_query(index=p_index, body=p_query, headers=headers)
        return l_result['hits'].get('total')

    def query_data(self, p_index, p_query, headers=None):
        """
        常用计数函数
        """
        l_result = self.base_query(index=p_index, body=p_query, headers=headers)
        return l_result['hits'].get('hits')

    def query_agg_data(self, p_index, p_query, p_agg_name='res', headers=None):
        """
        常用聚合函数
        """
        l_result = self.base_query(index=p_index, body=p_query, headers=headers)
        return l_result['aggregations'][p_agg_name]['buckets']

    def update(self, index, id, body, headers=None):
        """
        client = EsClient().update("test_log_1", 'xQAwR3YBGr1x55HFO65t', {'module': 'test_es'})
        """
        r_body = {'doc': body}
        self._client.update(index, id, r_body, headers=headers)

    def write(self, p_data_list):
        """
        数据写入函数

        """
        l_tmp_list = []
        for l_data in p_data_list:
            l_tmp_list.append(l_data)
            if len(l_tmp_list) == 1000:
                helpers.bulk(self._client, l_tmp_list)
                l_tmp_list.clear()

        if len(l_tmp_list) > 0:
            helpers.bulk(self._client, l_tmp_list)
            l_tmp_list.clear()

    def create_mapping(self, index_name, mapping):
        '''
        创建es mapping
        :param index_name:
        :param mapping:
        :return:
        '''
        result = ''
        if not self._client.indices.exists(index=index_name):
            self._client.indices.create(index=index_name, ignore=400)
        result = self._client.indices.put_mapping(
            index=index_name, body=mapping)
        return result

    def get_setting(self, index_name):
        '''
        获取索引设置
        :param index_name:
        :return:
        '''
        result = self._client.indices.get_settings(index_name)
        return result

    def get_mapping(self, index_name):
        '''
        获取索引mapping
        :param index_name:
        :return:
        '''
        result = self._client.indices.get_mapping(index_name)
        return result

    def put_settings(self, index_name, body):
        '''
        设置索引
        :param index_name:
        :param body:
        :return:
        '''
        result = self._client.indices.put_settings(body, index_name)
        return result

    def delete_index(self, index_name):
        '''
        删除索引
        :param index_name:
        :return:
        '''
        result = ''
        if self._client.indices.exists(index=index_name):
            result = self._client.indices.delete(
                index=index_name, ignore=[400, 404])
            return result
        return result

    def refresh(self, index_name):
        '''
        刷新索引
        :param index_name:
        :return:
        '''
        result = self._client.indices.refresh(index_name, request_timeout=600)
        return result

    def get_one_byid(self, index_name, _id):
        '''
        根据_id 查询数据
        :param index_name:
        :param _id:
        :return:
        '''
        result = self._client.get(index=index_name, id=_id)
        return result

    def query_all(self, index_name, query_dict):
        '''
        查询所有数据
        :param index_name:
        :param query_dict:
        :return:
        '''
        result = self._client.search(
            index=index_name,
            body=query_dict,
            request_timeout=600)
        return result

    def scroll_search(self, index_name, query_dict):
        '''
        针对大量数据使用scroll迭代查询
        :param index_name:
        :param query_dict:
        :return:
        '''
        result = self._client.search(
            index=index_name,
            body=query_dict,
            scroll="1m",
            request_timeout=600)
        return result

    def scroll(self, sid):
        '''
        根据scroll id查询
        :param sid:
        :return:
        '''
        result = self._client.scroll(
            scroll_id=sid, scroll='10m', request_timeout=600)
        return result

    def deleteAllDocByIndex(self, index_name):
        '''
        根据索引删除所有记录
        :param index_name:
        :return:
        '''
        query = {'query': {'match_all': {}}}
        return self._client.delete_by_query(
            index=index_name, body=query, request_timeout=600)

    def deleteAllByQuery(self, index_name, query):
        '''
        根据查询条件删除对应记录
        :param index_name:
        :param query:
        :return:
        '''
        return self._client.delete_by_query(index=index_name, body=query)

    def insert_one_byid(self, index_name, row_obj):
        '''
        根据_id 插入记录
        :param index_name:
        :param row_obj:
        :return:
        '''
        _id = row_obj.get("_id", 1)
        row_obj.pop("_id")
        result = self._client.index(index=index_name, body=row_obj, id=_id)
        return result

    def insert_one(self, index_name, body):
        '''
        插入一条记录
        :param index_name:
        :param body:
        :return:
        '''
        result = self._client.index(index=index_name, body=body)
        return result

    def update_data_bulk(self, index_name, row_obj_list, upsert=False):
        '''
        批量更新
        :param index_name: 索引名
        :param row_obj_list: 记录列表，每条需要有_id
        :return:
        '''
        load_data = []
        for row_obj in row_obj_list:
            # new
            action = {}
            if "_id" in row_obj:
                action['_id'] = row_obj.get('_id', 'None')
                row_obj.pop("_id")
                action['_op_type'] = 'update'
                action['doc'] = {}
                for r_o_i in row_obj:
                    action['doc'][r_o_i] = row_obj.get(r_o_i, None)
                if upsert:
                    action['doc_as_upsert'] = True
                load_data.append(action)
        result = {}
        if load_data:
            success, failed = helpers.bulk(
                self._client, load_data, index=index_name, raise_on_error=True, request_timeout=1000)
            result['success'] = success
            result['failed'] = failed
        return result

    def add_data_bulk(self, index_name, row_obj_list):
        '''
        批量插入
        :param index_name:
        :param row_obj_list:
        :return:
        '''
        load_data = []
        for row_obj in row_obj_list:
            # new
            action = {}
            action['_index'] = index_name
            if '_id' in row_obj:
                action['_id'] = row_obj.get('_id', 'None')
                row_obj.pop("_id")
            action['_source'] = {}
            for r_o_i in row_obj:
                action['_source'][r_o_i] = row_obj.get(r_o_i, None)
            load_data.append(action)
        result = {}
        if load_data:
            success, failed = helpers.bulk(
                self._client, load_data, index=index_name, raise_on_error=True, request_timeout=1000)
            result['success'] = success
            result['failed'] = failed
        return result

    def delete_data_bulk(self, index_name, row_obj_list):
        '''
        批量删除
        :param index_name:
        :param row_obj_list:
        :return:
        '''
        load_data = []
        for row_obj in row_obj_list:
            # new
            action = {}
            action['_op_type'] = 'delete'
            if "_id" in row_obj:
                action['_id'] = row_obj.get('_id', 'None')
                row_obj.pop("_id")
                load_data.append(action)
        result = {}
        if load_data:
            success, failed = helpers.bulk(
                self._client, load_data, index=index_name, raise_on_error=True, request_timeout=1000)
            result['success'] = success
            result['failed'] = failed
        return result

    def update_by_id(self, index_name, row_obj):
        '''
        根据_id更新记录
        :param index_name:
        :param row_obj:
        :return:
        '''
        _id = row_obj.get("_id", 1)
        row_obj.pop("_id")
        result = self._client.update(
            index=index_name, body={
                "doc": row_obj}, id=_id, request_timeout=600)
        return result

    def delete_by_id(self, index_name, _id):
        '''
        根据_id 删除记录
        :param index_name:
        :param _id:
        :return:
        '''
        result = self._client.delete(index=index_name, id=_id)
        return result

    def deleteDocByQuery(self, index_name, query):
        '''
        根据查询条件删除记录
        :param index_name:
        :param query:
        :return:
        '''
        return self._client.delete_by_query(index=index_name, body=query)




'''
es查询接口封装
'''
import requests
from ezetl.libs.es import EsClient
import re
from urllib import parse
import json
import base64


class EsQueryTool(object):
    '''
    es根据接口字段拼接query查询
    '''
    def __init__(self, params, index_names=[]):
        '''
        初始化参数
        :param params: 查询参数
        :param index_names: 允许查询索引名列表
        '''
        self.params = params
        self.params_lists = []
        for k in self.params:
            self.params_lists.append((k, [self.params[k]]))
        self.page = int(params.get('page', 1))
        self.pagesize = int(params.get('pagesize', 20))
        self.es_host = params.get('host')
        self.es_port = params.get('port')
        self.es_username = params.get('username')
        self.es_password = params.get('password')
        self.api_test = params.get('api_test', 0)
        if index_names == []:
            self.index_name = params.get('index_name', '')
        else:
            index_name = params.get('index_name', '')
            if index_name == '':
                self.index_name = ','.join(index_names)
            else:
                li = index_name.split(',')
                names = [i for i in li if i in index_names]
                self.index_name = ','.join(names)
        self.query_body = {
            'from': (self.page - 1) * self.pagesize,
            'size': self.pagesize,
            'query': {},
            'sort': {},
            'aggs': {},
            'track_total_hits': True  # 获取真实数量，不设置会只返回10000
        }
        self.content_tag = self.params.get('content_tag', '1')
        self.doc_ids = self.parse_doc_ids()
        self.equal_list = self.parse_equal()
        self.f_equal_list = self.parse_f_equal()
        self.contain_list = self.parse_contain()
        self.f_contain_list = self.parse_f_contain()
        self.range_params = self.parse_range()
        self.jl_tag = self.parse_jl_tag()
        self.search_key = self.parse_search_key()
        self.sort = self.parse_sort()
        self.return_fields = self.parse_return_fields()
        self.err_info = None

    def parse_jl_tag(self):
        '''
        解析聚类字段
        jl_tag[coin]=coin_bucket&jl_type[coin_bucket]=terms&jl_size[coin_bucket]=10
        :return:
        '''
        jl_tag = self.params.get('jl_tag', '')
        jl_pagesize_map = {}
        for k, li in self.params_lists:
            if k.startswith('jl_page_size'):
                key = re.findall('\[(.*?)\]', k)
                if key and key != ['']:
                    for i in li:
                        try:
                            jl_pagesize_map[key[0]] = int(i)
                        except Exception as e:
                            print(e)
        jl_type_map = {}
        for k, li in self.params_lists:
            if k.startswith('jl_type'):
                key = re.findall('\[(.*?)\]', k)
                if key and key != ['']:
                    for i in li:
                        jl_type_map[key[0]] = i
        jl_tag_list = []
        if jl_tag != '':
            jl_tag_list = jl_tag.split(',')
            for tag in jl_tag_list:
                pagesize = jl_pagesize_map.get(tag, 20)
                jl_type = jl_type_map.get(tag, 'terms')
                if jl_type == 'terms':
                    self.query_body['aggs'][tag] = {"terms": {"field": tag, "size": pagesize, "shard_size": 1200}}
                elif jl_type in ['max', 'min', 'avg', 'sum', 'stats']:
                    self.query_body['aggs'][tag] = {jl_type: {"field": tag}}
        return jl_tag_list

    def change_search_key(self, tmp_str):
        tmp_str = parse.unquote(str(tmp_str).strip())
        match = re.findall("[A-Za-z0-9\_]+[(](.*?)[)]", tmp_str)
        if match:
            for ii in match:
                if ii[0] == '"' and ii[-1] == '"':
                    tmp_str = tmp_str.replace('(' + ii + ')', ' = ' + ii, 1)
                else:
                    if ii[0] == '"':
                        tmp_str = tmp_str.replace('(' + ii + ')', ' like ' + ii[1:], 1)
                    elif ii[-1] == '"':
                        tmp_str = tmp_str.replace('(' + ii + ')', ' like ' + ii[:-1], 1)
                    else:
                        if ii == '^':
                            tmp_str = tmp_str.replace('(' + ii + ')', ' is not null ', 1)
                        else:
                            tmp_str = tmp_str.replace('(' + ii + ')', ' like ' + ii, 1)
        else:
            if '=' in tmp_str:
                tmp_str = tmp_str.replace('=', ' like ', 1)
            else:
                if tmp_str[0] == '"' and tmp_str[-1] == '"':
                    tmp_str = {"match_phrase": {"_all": {"query": tmp_str}}}
                else:
                    tmp_str = {"multi_match": {"query": tmp_str, "type": "most_fields", "fields": [], "operator": "and",
                                               "tie_breaker": 0.3, "minimum_should_match": "75%", "slop": 1}}
        return tmp_str

    def parse_sql(self, sql):
        '''
        解析sql语句拼接query
        :param sql:
        :return:
        '''
        try:
            if 'where' not in sql:
                sql = f'select * from {self.index_name} where {sql}'
            headers = {
                'Content-Type': 'application/json'
            }
            if self.es_username != '':
                headers['Authorization'] = f'Basic {base64.b64encode(f"{self.es_username}:{self.es_password}".encode()).decode()}'
            req_data = {
                'sql': sql
            }
            url = f'http://{self.es_host}:{str(self.es_port)}/_nlpcn/sql/explain'
            res = requests.post(url, headers=headers, json=req_data)
            res = json.loads(res.text)
            if 'error' in res:
                return True, res['error']

            if 'bool' not in self.query_body['query']:
                self.query_body['query']['bool'] = {}
            if 'must' not in self.query_body['query']['bool']:
                self.query_body['query']['bool']['must'] = []
            if 'bool' in res['query'] and 'filter' in res['query']['bool']:
                self.query_body['query']['bool']['must'].extend(res['query']['bool']['filter'])
            return False, res
        except Exception as e:
            return True, str(e)

    def parse_search_key(self):
        '''
        解析search_key参数
        :return:
        '''
        search_key = self.params.get('search_key', '')
        if search_key:
            if search_key.startswith('sql:'):  # 以sql: 开头的解析sql语句
                err, err_info = self.parse_sql(search_key[4:])
                if err:
                    self.err_info = err_info
            else:
                tmp_str = self.change_search_key(search_key)
                if isinstance(tmp_str, dict):
                    if 'bool' not in self.query_body['query']:
                        self.query_body['query']['bool'] = {}
                    if 'must' not in self.query_body['query']['bool']:
                        self.query_body['query']['bool']['must'] = []
                    self.query_body['query']['bool']['must'].append(tmp_str)
                else:
                    sql = f'select * from {self.index_name} where {tmp_str}'
                    err, err_info = self.parse_sql(sql)
                    if err:
                        self.err_info = err_info
        return search_key

    def parse_equal(self):
        '''
        解析equal参数, 根据聚类字段值筛选数据。
        普通字段 equal[key]=value => obj.key == value
        嵌套字段 equal[key1][key2] => obj.key1.key2 == value
        例：equal[PY][]=2000,2001，取PY为2000或者2001的数据
        :return:
        '''
        equal_list = []
        for k, li in self.params_lists:
            if k.startswith('equal'):
                key = re.findall('\[(.*?)\]', k)
                if key != ['']:
                    for i in li:
                        equal_list.append({'key': key, 'value': i.split(',')})
        if equal_list != []:
            if 'bool' not in self.query_body['query']:
                self.query_body['query']['bool'] = {}
            if 'filter' not in self.query_body['query']['bool']:
                self.query_body['query']['bool']['filter'] = []
            for equal in equal_list:
                if len(equal['key']) == 1 or len(equal['key']) == 2 and equal['key'][1] == '':  # 普通字段
                    terms = {'terms': {equal['key'][0]: equal['value']}}
                else:  # 嵌套字段
                    key1 = equal['key'][0]
                    key2 = equal['key'][1]
                    terms = {"nested": {"path": f"{key1}", "query": {
                        "bool": {"filter": [{"terms": {f"{key1}.{key2}": equal['value']}}]}}}}
                self.query_body['query']['bool']['filter'].append(terms)
        return equal_list

    def parse_f_equal(self):
        '''
        解析f_equal参数，用于排除不需要的数据
        普通字段 f_equal[key]=value => obj.key != value
        嵌套字段 f_equal[key1][key2]=value => obj.key1.key2 != value
        f_equal[PY][]=2000&f_equal[PY][]=2001或f_equal[PY][]=2000,2001  取PY不等于2000且不等于2001的数据
        :return:
        '''
        f_equal_list = []
        for k, li in self.params_lists:
            if k.startswith('f_equal'):
                key = re.findall('\[(.*?)\]', k)
                if key != ['']:
                    for i in li:
                        f_equal_list.append({'key': key, 'value': i.split(',')})
        if f_equal_list != []:
            if 'bool' not in self.query_body['query']:
                self.query_body['query']['bool'] = {}
            if 'must_not' not in self.query_body['query']['bool']:
                self.query_body['query']['bool']['must_not'] = []
            for f_equal in f_equal_list:
                if len(f_equal['key']) == 1 or len(f_equal['key']) == 2 and f_equal['key'][1] == '':  # 普通字段
                    terms = {'terms': {f_equal['key'][0]: f_equal['value']}}
                else:  # 嵌套字段
                    key1 = f_equal['key'][0]
                    key2 = f_equal['key'][1]
                    terms = {"nested": {"path": f"{key1}", "query": {
                        "bool": {"must": [{"terms": {f"{key1}.{key2}": f_equal['value']}}]}}}}
                self.query_body['query']['bool']['must_not'].append(terms)
        return f_equal_list

    def parse_contain(self):
        '''
        解析contain参数, 查询包含条件。
        普通字段 contain[key]=value => obj.key like value
        嵌套字段 contain[key1][key2] => obj.key1.key2 like value
        例：contain[TITLE][]= beijing 取TITLE中包含beijing的数据
        :return:
        '''
        contain_list = []
        for k, li in self.params_lists:
            if k.startswith('contain'):
                key = re.findall('\[(.*?)\]', k)
                if key != ['']:
                    for i in li:
                        contain_list.append({'key': key, 'value': i})
        if contain_list != []:
            if 'bool' not in self.query_body['query']:
                self.query_body['query']['bool'] = {}
            if 'must' not in self.query_body['query']['bool']:
                self.query_body['query']['bool']['must'] = []
            for contain in contain_list:
                if len(contain['key']) == 1 or len(contain['key']) == 2 and contain['key'][1] == '':  # 普通字段
                    terms = {'match_phrase': {contain['key'][0]: contain['value']}}
                else:  # 嵌套字段
                    key1 = contain['key'][0]
                    key2 = contain['key'][1]
                    terms = {"nested": {"path": f"{key1}", "query": {
                        "bool": {"must": [{"match_phrase": {f"{key1}.{key2}": [contain['value']]}}]}}}}
                self.query_body['query']['bool']['must'].append(terms)
        return contain_list

    def parse_f_contain(self):
        '''
        解析f_contain参数, 查询不包含条件。
        普通字段 f_contain[key]=value => obj.key like value
        嵌套字段 f_contain[key1][key2] => obj.key1.key2 like value
        例：f_contain[TITLE][]= beijing 取TITLE中不包含beijing的数据
        :return:
        '''
        contain_list = []
        for k, li in self.params_lists:
            if k.startswith('f_contain'):
                key = re.findall('\[(.*?)\]', k)
                if key != ['']:
                    for i in li:
                        contain_list.append({'key': key, 'value': i})
        if contain_list != []:
            if 'bool' not in self.query_body['query']:
                self.query_body['query']['bool'] = {}
            if 'must_not' not in self.query_body['query']['bool']:
                self.query_body['query']['bool']['must_not'] = []
            for contain in contain_list:
                if len(contain['key']) == 1 or len(contain['key']) == 2 and contain['key'][1] == '':  # 普通字段
                    terms = {'match_phrase': {contain['key'][0]: contain['value']}}
                else:  # 嵌套字段
                    key1 = contain['key'][0]
                    key2 = contain['key'][1]
                    terms = {"nested": {"path": f"{key1}", "query": {
                        "bool": {"must": [{"match_phrase": {f"{key1}.{key2}": [contain['value']]}}]}}}}
                self.query_body['query']['bool']['must_not'].append(terms)
        return contain_list

    def parse_range(self):
        '''
        解析range参数, gt, gte,lt, lte
        例：gt[PY]=2000&lt[PY]=2010表示取PY（2001~2009）,
        包含起始；一个参数 gte[PY]=2000表示PY大于等于2010
        :return:
        '''
        range_map = {}
        for k, li in self.params_lists:
            if k.startswith(('gt', 'gte', 'lt', 'lte')):
                key = re.findall('\[(.*?)\]', k)
                rule = k.split('[')[0]
                if key != [''] and len(li) == 1:
                    key = str(key)
                    if key not in range_map:
                        range_map[key] = {rule: li[0]}
                    else:
                        range_map[key][rule] = li[0]
        if range_map != {}:
            if 'bool' not in self.query_body['query']:
                self.query_body['query']['bool'] = {}
            if 'filter' not in self.query_body['query']['bool']:
                self.query_body['query']['bool']['filter'] = []
            for key, value in range_map.items():
                key = eval(key)
                for k in value:
                    value[k] = value[k]
                if len(key) == 1 or len(key) == 2 and key[1] == '':  # 普通字段
                    terms = {'range': {key[0]: [value]}}
                else:  # 嵌套字段
                    key1 = key[0]
                    key2 = key[1]
                    terms = {"nested": {"path": f"{key1}", "query": {
                        "bool": {"must": [{"range": {f"{key1}.{key2}": [value]}}]}}}}
                self.query_body['query']['bool']['filter'].append(terms)
        return range_map

    def parse_doc_ids(self):
        '''
        解析doc_ids参数, 根据唯一id获取内容，eg：1,2,3多个以逗号连接
        例：doc_ids=1,2,3表示取_id=1或2或3的值
        :return:
        '''
        doc_ids = self.params.get('doc_ids', '')
        if doc_ids != '':
            doc_ids = doc_ids.split(',')
            if 'bool' not in self.query_body['query']:
                self.query_body['query']['bool'] = {}
            if 'filter' not in self.query_body['query']['bool']:
                self.query_body['query']['bool']['filter'] = []
            terms = {'terms': {"_id": doc_ids}}
            self.query_body['query']['bool']['filter'].append(terms)
            return doc_ids
        else:
            return []

    def parse_return_fields(self):
        '''
        解析return_fields参数, 根据筛选返回字段。
        :return:
        '''
        return_fields = self.params.get('return_fields', '')
        if return_fields != '':
            return_fields = return_fields.split(',')
            self.query_body["_source"] = {
                "includes": return_fields
            }
        else:
            return_fields = []
        return return_fields

    def parse_sort(self):
        '''
        解析sort参数, 内容排序，eg:PY DESC,id ASC
        例：sort=PY DESC或sort=PY+DESC或sort[PY]=DESC 表示按PY从大到小排序，
        sort=id ASC或sort=id+ASC或sort[id]=ASC 表示按id从小到大排序
        :return:
        '''
        sort_list = []
        for k, li in self.params_lists:
            if k == 'sort':
                for i in li:
                    if '+' in i:
                        v = i.split('+')
                    else:
                        v = i.split(' ')
                    if len(v) == 2:
                        key, value = v
                        value = value.upper()
                        if value in ['DESC', 'ASC']:
                            sort_list.append({'key': key, 'value': value})
            if k.startswith('sort['):
                key = re.findall('\[(.*?)\]', k)
                if key != [''] and len(li) == 1:
                    key = key[0]
                    value = li[0].upper()
                    if value in ['DESC', 'ASC']:
                        sort_list.append({'key': key, 'value': value})
        for sort in sort_list:
            if sort['key'] not in self.query_body['sort']:
                self.query_body['sort'][sort['key']] = {"order": sort['value']}
        return sort_list

    def valid_params(self):
        '''
        检测各参数合法性
        :return:
        '''
        if self.page < 1:
            return {
                'code': 400,
                'msg': '页数不能小于1'
            }
        if self.index_name == '':
            return {
                'code': 400,
                'msg': '待查索引为空'
            }
        return False

    def gen_contents(self, data_li, valid_fields=['*']):
        '''
        根据content_tag类型组合返回内容
        '''
        content_tag = str(self.content_tag)
        if content_tag == '0':
            contents = []
        elif content_tag == '1':
            contents = []
            for i in data_li:
                if '*' not in valid_fields:
                    i['_source'] = {k: v for k, v in i['_source'].items() if k in valid_fields}
                dic = {
                    '_id': i['_id'],
                    '_index': i['_index'],
                    '_score': i['_score'],
                    '_type': i['_type'],
                    **i['_source']
                }
                contents.append(dic)
        elif content_tag == '2':
            contents = {}
            for i in data_li:
                if '*' not in valid_fields:
                    i['_source'] = {k: v for k, v in i['_source'].items() if k in valid_fields}
                source_dict = i['_source']
                for k, v in source_dict.items():
                    if k not in contents:
                        contents[k] = [v]
                    else:
                        contents[k].append(v)
        else:
            contents = []
            for i in data_li:
                if '*' not in valid_fields:
                    i['_source'] = {k: v for k, v in i['_source'].items() if k in valid_fields}
                contents.append(i)
        return contents

    def gen_result(self, result, valid_fields=['*']):
        '''
        组合返回结果
        :return:
        '''
        total = result['hits']['total']['value']
        data_li = result['hits']['hits']
        contents = self.gen_contents(data_li, valid_fields)
        aggregations = result.get('aggregations', {})
        if self.api_test != 0:
            res_data = {
                'data': {
                    'query_body': self.query_body,
                    'result': result,
                    'total': total,
                },
                'code': 200,
                'msg': '查询成功'
            }
        else:
            res_data = {
                'code': 200,
                'msg': '查询成功',
                'data': {
                    'records': contents,
                    'aggs': aggregations,
                    'total': total
                },
            }
        return res_data

    def query(self, valid_fields=['*'], es=None):
        '''
        执行查询返回结果
        :return:
        '''
        not_valid = self.valid_params()
        if not_valid:
            return not_valid
        if self.err_info:
            res_data = {
                'code': 400,
                'msg': '查询错误',
                'err_info': self.err_info
            }
            return res_data
        if self.query_body['query'] == {}:
            self.query_body['query'] = {'match_all': {}}
        try:
            if es is None:
                es_hosts = [{'host': self.es_host, 'port': self.es_port}]
                es = EsClient(hosts=es_hosts, http_auth=(self.es_username, self.es_password))
            print(self.query_body)
            result = es.query_all(self.index_name, self.query_body)
            res_data = self.gen_result(result, valid_fields)
            return res_data
        except Exception as e:
            print(e)
            return {
                'code': 400,
                'msg': f'接口错误{str(e)}'
            }


if __name__ == '__main__':
    params = {
        'page': 1,
        'pagesize': 20,
        'index_name': 'test',
        'gte[PY]': 2020,
        'lt[PY]': 2022,
        'equal[name]': '张三',
        'sort[PY]': 'ASC',
        'search_key': 'sql:select * from datacenter_logs where ip_raw is null'
    }
    es_tool = EsQueryTool(params=params)
    print(es_tool.query_body)



# -*- coding: utf-8 -*-
import logging
from ezetl.utils import get_reader, get_writer, get_res_fields
from ezetl.transform_algs import transform_alg_dict, df_to_data
from ezetl.utils.common_utils import gen_json_response


class EtlTask(object):
    def __init__(self, task_params):
        self.params = task_params
        self.extract = self.params.get('extract')
        self.extract_info = self.extract.get('extract_info', {})
        self.extract_type = self.extract_info.get('extract_type', 'once')
        self.process_rules = task_params.get('process_rules', [])
        self.load_info = self.params.get('load')
        self.error_list = []

    def gen_data_models(self):
        '''
        获取读取或写入数据模型
        :return:
        '''
        flag, reader = get_reader(self.extract)
        if flag:
            self.reader = reader
        else:
            self.error_list.append(reader)
            self.reader = None
        if self.load_info is not None:
            flag, writer = get_writer(self.load_info)
            if flag:
                self.writer = writer
            else:
                self.writer = None
        else:
            self.writer = None

    def process_batch(self, res_data, run_load=True):
        '''
        迭代生成器批量处理数据
        :param run_load: 是否执行数据装载
        :return:
        '''
        # 数据转换
        idx = 1
        context = {}
        for rule in self.process_rules:
            method_code = rule.get('code')
            alg_method = transform_alg_dict.get(method_code)
            if not alg_method:
                return False, f'数据转换第{idx}条规则出错：未找到处理函数{method_code}'
            rule_dict = rule.get('rule_dict')
            flag, res_data = alg_method(res_data, rule_dict, context)
            if not flag:
                return False, f'数据转换第{idx}条规则出错：{res_data}'
            idx += 1
        if run_load:
            if not self.writer:
                return False, f'数据装载出错：未找到数据装载对象'
            flag, res_data = self.writer.write(res_data)
            if not flag:
                return False, f'数据装载出错：{res_data}'
        return True, res_data

    def process_once(self, page=1, pagesize=20, run_load=True):
        '''
        单次处理数据
        :param page: 读取页数
        :param pagesize: 每页数量
        :param run_load: 是否执行数据装载
        :return:
        '''
        # 数据抽取
        flag, res_data = self.reader.read_page(page, pagesize)
        if not flag:
            return False, f"数据抽取出错：{res_data}"
        # 抽取成功，取出其中的数据
        res_data = res_data['data']
        # 数据转换
        idx = 1
        context = {}
        for rule in self.process_rules:
            method_code = rule.get('code')
            rule_dict = rule.get('rule_dict')
            alg_method = transform_alg_dict.get(method_code)
            if not alg_method:
                return False, f'数据转换第{idx}条规则出错：未找到处理函数{method_code}'
            flag, res_data = alg_method(res_data, rule_dict, context)
            if not flag:
                return False, f'数据转换第{idx}条规则出错：{res_data}'
            idx += 1
        if run_load:
            if not self.writer:
                return False, f'数据装载出错：未找到数据装载对象'
            flag, res_data = self.writer.write(res_data)
            if not flag:
                return False, f'数据装载出错：{res_data}'
        return True, '数据处理完成'

    def preview(self, run_load=False):
        '''
        处理数据预览模式，检查数据抽取转换流程
        :param run_load: 是否执行数据装载
        :return:
        '''
        if self.error_list != []:
            res_info = {
                'code': 400,
                'msg': f"任务初始化出错：{','.join(self.error_list)}"
            }
            return res_info
        # 数据抽取，抽取最多100条测试数据
        page_size = self.extract_info.get('batch_size', 100)
        if page_size > 100:
            page_size = 100
        flag, res_data = self.reader.read_page(pagesize=page_size)
        if not flag:
            res_info = {
                'code': 400,
                'msg': f"数据抽取出错：{res_data}"
            }
            return res_info
        # 抽取成功，取出其中的数据
        res_data = res_data['data']
        # 数据转换
        idx = 1
        context = {}
        for rule in self.process_rules:
            method_code = rule.get('code')
            rule_dict = rule.get('rule_dict')
            alg_method = transform_alg_dict.get(method_code)
            if not alg_method:
                res_info = {
                    'code': 400,
                    'msg': f'数据转换第{idx}条规则出错：未找到处理函数{method_code}'
                }
                return res_info
            flag, res_data = alg_method(res_data, rule_dict, context)
            if not flag:
                res_info = {
                    'code': 400,
                    'msg': f'数据转换第{idx}条规则出错：{res_data}'
                }
                return res_info
            idx += 1
        if run_load:
            if not self.writer:
                res_info = {
                    'code': 400,
                    'msg': f'数据装载出错：未找到数据装载对象'
                }
                return res_info
            flag, res_data = self.writer.write(res_data)
            if not flag:
                res_info = {
                    'code': 400,
                    'msg': f'数据装载出错：{res_data}'
                }
                return res_info
        # 若是dataframe，转回json数据用于前端展示
        flag, res_data = df_to_data(res_data)
        if not flag:
            res_info = {
                'code': 400,
                'msg': f'{res_data}'
            }
            return res_info
        # 获取返回字段
        res_fields = get_res_fields(res_data)
        # 获取数据筛选规则列表
        extract_rules = self.reader.get_extract_rules()
        search_type_list = self.reader.get_search_type_list()
        res_data = {
            'data': res_data,
            'fields': res_fields,
            'extract_rules': extract_rules,
            'search_type_list': search_type_list,
        }
        return gen_json_response(data=res_data)


def etl_task_process(params, run_load=False, logger=None, task_class=EtlTask):
    '''
    执行处理任务
    :return:
    '''
    if logger is None:
        logger = logging.getLogger(__name__)
    etl_task = task_class(params)
    etl_task.gen_data_models()
    if etl_task.error_list != []:
        logger.error(f"任务初始化出错：{','.join(etl_task.error_list)}")
        return
    batch_size = etl_task.reader.batch_size
    extract_type = etl_task.extract_type
    if extract_type == 'once':
        flag, res_data = etl_task.process_once(pagesize=batch_size)
        if not flag:
            logger.error(f'数据装载出错：{res_data}')
    elif extract_type in ['batch', 'flow']:
        # 数据抽取
        reader_gen = etl_task.reader.read_batch()
        while reader_gen:
            try:
                flag, res_data = next(reader_gen)
                if not flag:
                    logger.error(f"数据抽取出错：{res_data}")
                    break
                # 抽取成功，取出其中的数据
                res_data = res_data['data']
                flag, res_data = etl_task.process_batch(res_data, run_load=run_load)
                if not flag:
                    logger.error(f"数据处理出错：{res_data}")
                    break
                if run_load:
                    if not etl_task.writer:
                        res_info = {
                            'code': 400,
                            'msg': f'数据装载出错：未找到数据装载对象'
                        }
                        return res_info
                    flag, res_data = etl_task.writer.write(res_data)
                    if not flag:
                        res_info = {
                            'code': 400,
                            'msg': f'数据装载出错：{res_data}'
                        }
                        return res_info
            except StopIteration:
                logger.info('数据处理完成')
                break
            except Exception as e:
                logger.exception(e)
                break











'''
使用任务配置字典分批读取mysql，内置算法聚合统计数据，写入目标kafka数据源
'''
from ezetl.etl_task import etl_task_process

task_params = {
    'extract': {
        'source': {
            "name": "test",
            "type": "mysql",
            "conn_conf": {
                "host": "127.0.0.1",
                "port": 3306,
                "username": "root",
                "password": "123456",
                "database_name": "test"
            },
            "ext_params": {}
        },
        'model': {
            "name": "btc_history",
            "type": "mysql_table",
            "model_conf": {
                "name": "btc_history",
            },
            "ext_params": {},
            "fields": []
        },
        'extract_info': {
            'batch_size': 1000,
            'extract_rules': []
        }
    },
    'process_rules': [
        {
            "code": "gen_records_list",
            "name": "获取内容列表",
            "rule_dict": {
                "fields": "time,symbol,close"
            }
        },
        {
            "code": "trans_time_format",
            "name": "转换时间格式",
            "rule_dict": {
                "fields": "time",
                "format": "%Y-%m"
            }
        },
        {
            "code": "trans_field_type",
            "name": "转换字段类型",
            "rule_dict": {
                "fields": "close",
                "trans_type": "float"
            }
        },
        {
            "code": "group_agg_count",
            "name": "分组聚合统计",
            "rule_dict": {
                "count_field": "close",
                "group_fields": "time",
                "count_type": "mean"
            }
        },
        {
            "code": "df_to_data",
            "name": "dataframe转回原始数据",
            "rule_dict": {
            }
        }
    ],
    'load': {
        'source': {
            "name": "test",
            "type": "kafka",
            "conn_conf": {
                "bootstrap_servers": "127.0.0.1:9092",
            },
            "ext_params": {}
        },
        'model': {
            "name": "test",
            "type": "kafka_topic",
            "model_conf": {
                "name": "btc_history",
            },
            "ext_params": {},
            "fields": []
        },
        'load_info': {
            'load_type': 'insert',
            'only_fields': []
        }
    }
}
etl_task_process(task_params, run_load=True)
